# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
End-to-end memory network described in
(Sukhbaatar et al.) End-To-End Memory Networks
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from texar.module_base import ModuleBase
from texar.modules.embedders import WordEmbedder, PositionEmbedder
from texar.utils.mode import switch_dropout
from texar.modules.memory.embed_fns import default_memnet_embed_fn_hparams

# pylint: disable=invalid-name, too-many-instance-attributes, too-many-arguments
# pylint: disable=too-many-locals

__all__ = [
    'MemNetBase',
    'MemNetRNNLike',
]

class MemNetSingleLayer(ModuleBase):
    """An A-C layer for memory network.

    Args:
        H (optional): The matrix :attr:`H` multiplied to :attr:`o` at the end.
        hparams (dict or HParams, optional): Memory network single layer
            hyperparameters. If it is not specified, the default hyperparameter
            setting is used. See :attr:`default_hparams` for the structure and
            default values.
    """

    def __init__(self, H=None, hparams=None):
        ModuleBase.__init__(self, hparams)

        self._H = H

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "name": "memnet_single_layer"
            }

        Here:

        "name": str
            Name of the memory network single layer.
        """
        return {
            "name": "memnet_single_layer"
        }

    def _build(self, u, m, c, **kwargs):
        """An A-C operation with memory and query vector.

        Args:
            u (Tensor): The input query `Tensor` of shape `[None, memory_dim]`.
            m (Tensor): Output of A operation. Should be in shape
                `[None, memory_size, memory_dim]`.
            c (Tensor): Output of C operation. Should be in shape
                `[None, memory_size, memory_dim]`.

        Returns:
            A `Tensor` of shape same as :attr:`u`.
        """
        # Input memory representation
        p = tf.matmul(m, tf.expand_dims(u, axis=2))
        p = tf.transpose(p, perm=[0, 2, 1])

        p = tf.nn.softmax(p) # equ. (1)

        # Output memory representation
        o = tf.matmul(p, c) # equ. (2)
        o = tf.squeeze(o, axis=[1])

        if self._H:
            u = tf.matmul(u, self._H) # RNN-like style
        u_ = tf.add(u, o) # u^{k+1} = H u^k + o^k

        if not self._built:
            self._add_internal_trainable_variables()
            if self._H:
                self._add_trainable_variable(self._H)
            self._built = True

        return u_

class MemNetBase(ModuleBase):
    """Base class inherited by all memory network classes.

    Args:
        raw_memory_dim (int): Dimension size of raw memory entries
            (before embedding). For example,
            if a raw memory entry is a word, this is the **vocabulary size**
            (imagine a one-hot representation of word). If a raw memory entry
            is a dense vector, this is the dimension size of the vector.
        input_embed_fn (optional): A callable that embeds raw memory entries
            as inputs.
            This corresponds to the `A` embedding operation in
            (Sukhbaatar et al.)
            If not provided, a default embedding operation is created as
            specified in :attr:`hparams`. See
            :meth:`~texar.modules.MemNetBase.get_default_embed_fn`
            for details.
        output_embed_fn (optional): A callable that embeds raw memory entries
            as outputs.
            This corresponds to the `C` embedding operation in
            (Sukhbaatar et al.)
            If not provided, a default embedding operation is created as
            specified in :attr:`hparams`. See
            :meth:`~texar.modules.MemNetBase.get_default_embed_fn`
            for details.
        query_embed_fn (optional): A callable that embeds query.
            This corresponds to the `B` embedding operation in
            (Sukhbaatar et al.). If not provided and "use_B" is True
            in :attr:`hparams`, a default embedding operation is created as
            specified in :attr:`hparams`. See
            :meth:`~texar.modules.MemNetBase.get_default_embed_fn`
            for details.
            Notice: If you'd like to customize this callable, please follow
            the same number and style of dimensions as in `input_embed_fn` or
            `output_embed_fn`, and assume that the 2nd dimension of its
            input and output (which corresponds to `memory_size`) is 1.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparamerter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter sturcture and
            default values.
    """

    def __init__(self,
                 raw_memory_dim,
                 input_embed_fn=None,
                 output_embed_fn=None,
                 query_embed_fn=None,
                 hparams=None):
        ModuleBase.__init__(self, hparams)

        self._raw_memory_dim = raw_memory_dim

        self._n_hops = self._hparams.n_hops
        self._relu_dim = self._hparams.relu_dim
        self._memory_size = self._hparams.memory_size

        with tf.variable_scope(self.variable_scope):
            self._A, self._C, self._B, self._memory_dim = self._build_embed_fn(
                input_embed_fn, output_embed_fn, query_embed_fn)

            self.H = None
            if self.hparams.use_H:
                self.H = tf.get_variable(
                    name="H", shape=[self._memory_dim, self._memory_dim])

    def _build_embed_fn(self, input_embed_fn, output_embed_fn, query_embed_fn):
        # Optionally creates embed_fn's
        memory_dim = self.hparams.memory_dim
        mdim_A, mdim_C, mdim_B = None, None, None

        A = input_embed_fn
        if input_embed_fn is None:
            A, mdim_A = self.get_default_embed_fn(
                self._memory_size, self._hparams.A)
            memory_dim = mdim_A

        C = output_embed_fn
        if output_embed_fn is None:
            C, mdim_C = self.get_default_embed_fn(
                self._memory_size, self._hparams.C)
            if mdim_A is not None and mdim_A != mdim_C:
                raise ValueError('Embedding config `A` and `C` must have '
                                 'the same output dimension.')
            memory_dim = mdim_C

        B = query_embed_fn
        if query_embed_fn is None and self._hparams.use_B:
            B, mdim_B = self.get_default_embed_fn(1, self._hparams.B)
            if mdim_A is not None and mdim_A != mdim_B:
                raise ValueError('Embedding config `A` and `B` must have '
                                 'the same output dimension.')
            if mdim_C is not None and mdim_C != mdim_B:
                raise ValueError('Embedding config `C` and `B` must have '
                                 'the same output dimension.')
            memory_dim = mdim_B

        return A, C, B, memory_dim


    def get_default_embed_fn(self, memory_size, embed_fn_hparams):
        """Creates a default embedding function. Can be used for A, C, or B
        operation.

        For B operation (i.e., query_embed_fn), :attr:`memory_size` must be 1.

        The function is a combination of both memory embedding and temporal
        embedding, with the combination method specified by "combine_mode" in
        the `embed_fn_hparams`.

        .. role:: python(code)
           :language: python

        Args:
            embed_fn_hparams (dict or HParams): Hyperparameter of the
                embedding function. See
                :func:`~texar.modules.default_memnet_embed_fn` for details.

        Returns:
            A tuple `(embed_fn, memory_dim)`, where

            - **`memory_dim`** is the dimension of memory entry embedding, \
            inferred from :attr:`embed_fn_hparams`.

                - If `combine_mode` == 'add', `memory_dim` is the \
                embedder dimension.
                - If `combine_mode` == 'concat', `memory_dim` is the sum \
                of the memory embedder dimension and the temporal embedder \
                dimension.

            - **`embed_fn`** is an embedding function that takes in memory \
            and returns memory embedding. \
            Specifically, the function has signature \
            :python:`memory_embedding= embed_fn(memory=None, soft_memory=None)`\
            where one of `memory` and `soft_memory` is provided (but not both).

            Args:
                memory: An `int` Tensor of shape
                    `[batch_size, memory_size]`
                    containing memory indexes used for embedding lookup.
                soft_memory: A Tensor of shape
                    `[batch_size, memory_size, raw_memory_dim]`
                    containing soft weights used to mix the embedding vectors.

            Returns:
                A Tensor of shape `[batch_size, memory_size, memory_dim]`
                containing the memory entry embeddings.

        """
        # memory embedder
        embedder = WordEmbedder(
            vocab_size=self._raw_memory_dim,
            hparams=embed_fn_hparams["embedding"]
        )
        # temporal embedder
        temporal_embedder = PositionEmbedder(
            position_size=memory_size,
            hparams=embed_fn_hparams["temporal_embedding"]
        )

        combine = embed_fn_hparams['combine_mode']
        if combine == 'add':
            if embedder.dim != temporal_embedder.dim:
                raise ValueError('`embedding` and `temporal_embedding` must '
                                 'have the same dimension for "add" '
                                 'combination.')
            memory_dim = embedder.dim
        elif combine == 'concat':
            memory_dim = embedder.dim + temporal_embedder.dim

        def _embed_fn(memory, soft_memory, mode=None):
            if memory is None and soft_memory is None:
                raise ValueError(
                    "Either `memory` or `soft_memory` is required.")
            if memory is not None and soft_memory is not None:
                raise ValueError(
                    "Must not specify `memory` and `soft_memory` at the "
                    "same time.")

            embedded_memory = embedder(
                ids=memory, soft_ids=soft_memory, mode=mode)
            temporal_embedded = temporal_embedder(
                sequence_length=tf.constant([memory_size]), mode=mode)
            temporal_embedded = tf.tile(
                temporal_embedded, [tf.shape(embedded_memory)[0], 1, 1])

            if combine == 'add':
                return tf.add(embedded_memory, temporal_embedded)
            elif combine == 'concat':
                return tf.concat([embedded_memory, temporal_embedded], axis=-1)
            else:
                raise ValueError('Unknown combine method: {}'.format(combine))

        return _embed_fn, memory_dim

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "n_hops": 1,
                "memory_dim": 100,
                "relu_dim": 50,
                "memory_size": 100,
                "A": default_embed_fn_hparams,
                "C": default_embed_fn_hparams,
                "B": default_embed_fn_hparams,
                "use_B": False,
                "use_H": False,
                "dropout_rate": 0,
                "variational": False,
                "name": "memnet",
            }

        Here:

        "n_hops" : int
            Number of hops.

        "memory_dim" : int
            Memory dimension, i.e., the dimension size of a memory entry
            embedding. Ignored if at least one of the embedding functions is
            created according to :attr:`hparams`. In this case
            :attr:`memory_dim` is inferred from the created embed_fn.

        "relu_dim" : int
            Number of elements in :attr:`memory_dim` that have relu at the end
            of each hop.
            Should be not less than 0 and not more than :attr`memory_dim`.

        "memory_size" : int
            Number of entries in memory.

            For example, the number of sentences {x_i} in Fig.1(a) of
            (Sukhbaatar et al.) End-To-End Memory Networks.

        "use_B" : bool
            Whether to create the query embedding function. Ignored if
            `query_embed_fn` is given to the constructor.

        "use_H" : bool
            Whether to perform a linear transformation with matrix `H` at
            the end of each A-C layer.

        "dropout_rate" : float
            The dropout rate to apply to the output of each hop. Should
            be between 0 and 1.
            E.g., `dropout_rate=0.1` would drop out 10% of the units.

        "variational" : bool
            Whether to share dropout masks after each hop.
        """
        return {
            "n_hops": 1,
            "memory_dim": 100,
            "relu_dim": 50,
            "memory_size": 100,
            "A": default_memnet_embed_fn_hparams(),
            "C": default_memnet_embed_fn_hparams(),
            "B": default_memnet_embed_fn_hparams(),
            "use_B": False,
            "use_H": False,
            "dropout_rate": 0,
            "variational": False,
            "name": "memnet",
        }

    def _build(self, memory, query, **kwargs):
        raise NotImplementedError

    @property
    def memory_size(self):
        """The memory size.
        """
        return self._memory_size

    @property
    def raw_memory_dim(self):
        """The dimension of memory element (or vocabulary size).
        """
        return self._raw_memory_dim

    @property
    def memory_dim(self):
        """The dimension of embedded memory and all vectors in hops.
        """
        return self._memory_dim


class MemNetRNNLike(MemNetBase):
    """An implementation of multi-layer end-to-end memory network,
    with RNN-like weight tying described in
    (Sukhbaatar et al.) End-To-End Memory Networks .

    See :meth:`~texar.modules.MemNetBase.get_default_embed_fn` for default
    embedding functions. Customized embedding functions must follow
    the same signature.

    Args:
        raw_memory_dim (int): Dimension size of raw memory entries
            (before embedding). For example,
            if a raw memory entry is a word, this is the **vocabulary size**
            (imagine a one-hot representation of word). If a raw memory entry
            is a dense vector, this is the dimension size of the vector.
        input_embed_fn (optional): A callable that embeds raw memory entries
            as inputs.
            This corresponds to the `A` embedding operation in
            (Sukhbaatar et al.)
            If not provided, a default embedding operation is created as
            specified in :attr:`hparams`. See
            :meth:`~texar.modules.MemNetBase.get_default_embed_fn`
            for details.
        output_embed_fn (optional): A callable that embeds raw memory entries
            as outputs.
            This corresponds to the `C` embedding operation in
            (Sukhbaatar et al.)
            If not provided, a default embedding operation is created as
            specified in :attr:`hparams`. See
            :meth:`~texar.modules.MemNetBase.get_default_embed_fn`
            for details.
        query_embed_fn (optional): A callable that embeds query.
            This corresponds to the `B` embedding operation in
            (Sukhbaatar et al.). If not provided and "use_B" is True
            in :attr:`hparams`, a default embedding operation is created as
            specified in :attr:`hparams`. See
            :meth:`~texar.modules.MemNetBase.get_default_embed_fn`
            for details.
            For customized query_embed_fn, note that the function must follow
            the signature of the default embed_fn where `memory_size` must
            be 1.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparamerter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter sturcture and
            default values.
    """

    def __init__(self,
                 raw_memory_dim,
                 input_embed_fn=None,
                 output_embed_fn=None,
                 query_embed_fn=None,
                 hparams=None):
        MemNetBase.__init__(self, raw_memory_dim, input_embed_fn,
                            output_embed_fn, query_embed_fn, hparams)

        with tf.variable_scope(self.variable_scope):
            self._AC = MemNetSingleLayer(
                self.H,
                hparams={"name": "AC"})

            self._W = tf.layers.Dense(
                units=raw_memory_dim,
                use_bias=False,
                name="W")

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "n_hops": 1,
                "memory_dim": 100,
                "relu_dim": 50,
                "memory_size": 100,
                "A": default_embed_fn_hparams,
                "C": default_embed_fn_hparams,
                "B": default_embed_fn_hparams,
                "use_B": False,
                "use_H": True,
                "dropout_rate": 0,
                "variational": False,
                "name": "memnet_rnnlike",
            }

        Here:

        "n_hops" : int
            Number of hops.

        "memory_dim" : int
            Memory dimension, i.e., the dimension size of a memory entry
            embedding. Ignored if at least one of the embedding functions is
            created according to :attr:`hparams`. In this case
            :attr:`memory_dim` is inferred from the created embed_fn.

        "relu_dim" : int
            Number of elements in :attr:`memory_dim` that have relu at the end
            of each hop.
            Should be not less than 0 and not more than :attr`memory_dim`.

        "memory_size" : int
            Number of entries in memory.

            For example, the number of sentences {x_i} in Fig.1(a) of
            (Sukhbaatar et al.) End-To-End Memory Networks.

        "use_B" : bool
            Whether to create the query embedding function. Ignored if
            `query_embed_fn` is given to the constructor.

        "use_H" : bool
            Whether to perform a linear transformation with matrix `H` at
            the end of each A-C layer.

        "dropout_rate" : float
            The dropout rate to apply to the output of each hop. Should
            be between 0 and 1.
            E.g., `dropout_rate=0.1` would drop out 10% of the units.

        "variational" : bool
            Whether to share dropout masks after each hop.
        """
        hparams = MemNetBase.default_hparams()
        hparams.update({
            "use_H": True,
            "name": "memnet_rnnlike"
        })
        return hparams

    def _build(self, memory=None, query=None, soft_memory=None, soft_query=None,
               mode=None, **kwargs):
        """Pass the :attr:`memory` and :attr:`query` through the memory network
        and return the :attr:`logits` after the final matrix.

        Only one of :attr:`memory` and :attr:`soft_memory` can be specified.
        They should not be specified at the same time.

        Args:
            memory (optional): Memory used in A/C operations. By default, it
                should be an integer tensor of shape
                `[batch_size, memory_size]`,
                containing the ids to embed if provided.
            query (optional): Query vectors as the intial input of the memory
                network.
                If you'd like to apply some transformation (e.g., embedding)
                on it before it's fed into the network, please set `use_B` to
                True and add `query_embed_fn` when constructing this instance.
                If `query_embed_fn` is set to
                :meth:`~texar.modules.MemNetBase.get_default_embed_fn`,
                it should be of shape `[batch_size]`.
                If `use_B` is not set, it should be of shape
                `[batch_size, memory_dim]`.
            soft_memory (optional): Soft memory used in A/C operations. By
                default, it should be a tensor of shape
                `[batch_size, memory_size, raw_memory_dim]`,
                containing the weights used to mix the embedding vectors.
                If you'd like to apply a matrix multiplication on the memory,
                this option can also be used.
            soft_query (optional): Query vectors as the intial input of the
                memory network.
                If you'd like to apply some transformation (e.g., embedding)
                on it before it's fed into the network, please set `use_B` to
                True and add `query_embed_fn` when constructing this instance.
                Similar to :attr:`soft_memory`, if `query_embed_fn` is set to
                :meth:`~texar.modules.MemNetBase.get_default_embed_fn`,
                then it must be of shape `[batch_size, raw_memory_dim]`.
                Ignored if `use_B` is not set.
            mode (optional): A tensor taking value in
                :tf_main:`tf.estimator.ModeKeys <estimator/ModeKeys>`, including
                `TRAIN`, `EVAL`, and `PREDICT`. If `None`, dropout is
                controlled by :func:`texar.global_mode`.
        """
        if self._B is not None:
            def _unsqueeze(x):
                return x if x is None else tf.expand_dims(x, 1)
            query = tf.squeeze(
                self._B(_unsqueeze(query), _unsqueeze(soft_query), mode=mode),
                1)
        self._u = [query]
        self._m = self._A(memory, soft_memory, mode=mode)
        self._c = self._C(memory, soft_memory, mode=mode)

        keep_prob = switch_dropout(1-self.hparams.dropout_rate, mode=mode)
        if self.hparams.variational:
            with tf.variable_scope("variational_dropout"):
                noise = tf.random_uniform(tf.shape(self._u[-1]))
                random_tensor = keep_prob + noise
                binary_tensor = tf.floor(random_tensor)
            def _variational_dropout(val):
                return tf.div(val, keep_prob) * binary_tensor

        for _ in range(self._n_hops):
            u_ = self._AC(self._u[-1], self._m, self._c)
            if self._relu_dim == 0:
                pass
            elif self._relu_dim == self._memory_dim:
                u_ = tf.nn.relu(u_)
            elif 0 < self._relu_dim < self._memory_dim:
                linear_part = u_[:, : self._memory_dim - self._relu_dim]
                relu_part = u_[:, self._memory_dim - self._relu_dim :]
                relued_part = tf.nn.relu(relu_part)
                u_ = tf.concat(axis=1, values=[linear_part, relued_part])
            else:
                raise ValueError(
                    "relu_dim = {} is illegal".format(self._relu_dim))
            if self.hparams.variational:
                u_ = _variational_dropout(u_)
            else:
                u_ = tf.nn.dropout(u_, keep_prob)
            self._u.append(u_)

        logits = self._W(self._u[-1])

        if not self._built:
            self._add_internal_trainable_variables()
            self._built = True

        return logits



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various utilities specific to data processing.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import tarfile
import zipfile
import collections
import numpy as np
from six.moves import urllib
import requests

import tensorflow as tf

from texar.utils import utils_io

# pylint: disable=invalid-name, too-many-branches

__all__ = [
    "maybe_download",
    "read_words",
    "make_vocab",
    "count_file_lines"
]

Py3 = sys.version_info[0] == 3

def maybe_download(urls, path, filenames=None, extract=False):
    """Downloads a set of files.

    Args:
        urls: A (list of) urls to download files.
        path (str): The destination path to save the files.
        filenames: A (list of) strings of the file names. If given,
            must have the same length with :attr:`urls`. If `None`,
            filenames are extracted from :attr:`urls`.
        extract (bool): Whether to extract compressed files.

    Returns:
        A list of paths to the downloaded files.
    """
    utils_io.maybe_create_dir(path)

    if not isinstance(urls, (list, tuple)):
        urls = [urls]
    if filenames is not None:
        if not isinstance(filenames, (list, tuple)):
            filenames = [filenames]
        if len(urls) != len(filenames):
            raise ValueError(
                '`filenames` must have the same number of elements as `urls`.')

    result = []
    for i, url in enumerate(urls):
        if filenames is not None:
            filename = filenames[i]
        elif 'drive.google.com' in url:
            filename = _extract_google_drive_file_id(url)
        else:
            filename = url.split('/')[-1]
            # If downloading from GitHub, remove suffix ?raw=True
            # from local filename
            if filename.endswith("?raw=true"):
                filename = filename[:-9]

        filepath = os.path.join(path, filename)
        result.append(filepath)

        if not tf.gfile.Exists(filepath):
            if 'drive.google.com' in url:
                filepath = _download_from_google_drive(url, filename, path)
            else:
                filepath = _download(url, filename, path)

            if extract:
                tf.logging.info('Extract %s', filepath)
                if tarfile.is_tarfile(filepath):
                    tarfile.open(filepath, 'r').extractall(path)
                elif zipfile.is_zipfile(filepath):
                    with zipfile.ZipFile(filepath) as zfile:
                        zfile.extractall(path)
                else:
                    tf.logging.info("Unknown compression type. Only .tar.gz, "
                                    ".tar.bz2, .tar, and .zip are supported")

    return result

def _download(url, filename, path):
    def _progress(count, block_size, total_size):
        percent = float(count * block_size) / float(total_size) * 100.
        # pylint: disable=cell-var-from-loop
        sys.stdout.write('\r>> Downloading %s %.1f%%' %
                         (filename, percent))
        sys.stdout.flush()

    filepath = os.path.join(path, filename)
    filepath, _ = urllib.request.urlretrieve(url, filepath, _progress)
    print()
    statinfo = os.stat(filepath)
    print('Successfully downloaded {} {} bytes.'.format(
        filename, statinfo.st_size))

    return filepath

def _extract_google_drive_file_id(url):
    # id is between `/d/` and '/'
    url_suffix = url[url.find('/d/')+3:]
    file_id = url_suffix[:url_suffix.find('/')]
    return file_id

def _download_from_google_drive(url, filename, path):
    """Adapted from `https://github.com/saurabhshri/gdrive-downloader`
    """
    def _get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value
        return None

    file_id = _extract_google_drive_file_id(url)

    gurl = "https://docs.google.com/uc?export=download"
    sess = requests.Session()
    response = sess.get(gurl, params={'id': file_id}, stream=True)
    token = _get_confirm_token(response)

    if token:
        params = {'id': file_id, 'confirm': token}
        response = sess.get(gurl, params=params, stream=True)

    filepath = os.path.join(path, filename)
    CHUNK_SIZE = 32768
    with tf.gfile.GFile(filepath, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:
                f.write(chunk)

    print('Successfully downloaded {}.'.format(filename))

    return filepath

def read_words(filename, newline_token=None):
    """Reads word from a file.

    Args:
        filename (str): Path to the file.
        newline_token (str, optional): The token to replace the original newline
            token "\\\\n". For example,
            `newline_token=tx.data.SpecialTokens.EOS`.
            If `None`, no replacement is performed.

    Returns:
        A list of words.
    """
    with tf.gfile.GFile(filename, "r") as f:
        if Py3:
            if newline_token is None:
                return f.read().split()
            else:
                return f.read().replace("\n", newline_token).split()
        else:
            if newline_token is None:
                return f.read().decode("utf-8").split()
            else:
                return (f.read().decode("utf-8")
                        .replace("\n", newline_token).split())


def make_vocab(filenames, max_vocab_size=-1, newline_token=None,
               return_type="list", return_count=False):
    """Builds vocab of the files.

    Args:
        filenames (str): A (list of) files.
        max_vocab_size (int): Maximum size of the vocabulary. Low frequency
            words that exceeding the limit will be discarded.
            Set to `-1` (default) if no truncation is wanted.
        newline_token (str, optional): The token to replace the original newline
            token "\\\\n". For example,
            `newline_token=tx.data.SpecialTokens.EOS`.
            If `None`, no replacement is performed.
        return_type (str): Either "list" or "dict". If "list" (default), this
            function returns a list of words sorted by frequency. If "dict",
            this function returns a dict mapping words to their index sorted
            by frequency.
        return_count (bool): Whether to return word counts. If `True` and
            :attr:`return_type` is "dict", then a count dict is returned, which
            is a mapping from words to their frequency.

    Returns:
        - If :attr:`return_count` is False, returns a list or dict containing \
        the vocabulary words.

        - If :attr:`return_count` if True, returns a pair of list or dict \
        `(a, b)`, where `a` is a list or dict containing the vocabulary \
        words, `b` is a list of dict containing the word counts.
    """
    if not isinstance(filenames, (list, tuple)):
        filenames = [filenames]

    words = []
    for fn in filenames:
        words += read_words(fn, newline_token=newline_token)

    counter = collections.Counter(words)
    count_pairs = sorted(counter.items(), key=lambda x: (-x[1], x[0]))

    words, counts = list(zip(*count_pairs))
    if max_vocab_size >= 0:
        words = words[:max_vocab_size]
        counts = counts[:max_vocab_size]

    if return_type == "list":
        if not return_count:
            return words
        else:
            return words, counts
    elif return_type == "dict":
        word_to_id = dict(zip(words, range(len(words))))
        if not return_count:
            return word_to_id
        else:
            word_to_count = dict(zip(words, counts))
            return word_to_id, word_to_count
    else:
        raise ValueError("Unknown return_type: {}".format(return_type))


def count_file_lines(filenames):
    """Counts the number of lines in the file(s).
    """
    def _count_lines(fn):
        with open(fn, "rb") as f:
            i = -1
            for i, _ in enumerate(f):
                pass
            return i + 1

    if not isinstance(filenames, (list, tuple)):
        filenames = [filenames]
    num_lines = np.sum([_count_lines(fn) for fn in filenames])
    return num_lines




# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions related to tensor shapes.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

# pylint: disable=no-name-in-module, protected-access, no-member, invalid-name

import numpy as np

import tensorflow as tf
from tensorflow.python.util import nest
from tensorflow.python.ops import rnn
from tensorflow.python.framework import ops

__all__ = [
    "transpose_batch_time",
    "get_batch_size",
    "get_rank",
    "mask_sequences",
    "_mask_sequences_tensor",
    "_mask_sequences_py",
    "flatten",
    "shape_list",
    "pad_and_concat"
]


def transpose_batch_time(inputs):
    """Transposes inputs between time-major and batch-major.

    Args:
        inputs: A Tensor of shape `[batch_size, max_time, ...]` (batch-major)
            or `[max_time, batch_size, ...]` (time-major), or a (possibly
            nested) tuple of such elements.

    Returns:
        A (possibly nested tuple of) Tensor with transposed batch and
        time dimensions of inputs.
    """
    flat_input = nest.flatten(inputs)
    flat_input = [ops.convert_to_tensor(input_) for input_ in flat_input]
    # pylint: disable=protected-access
    flat_input = [rnn._transpose_batch_time(input_) for input_ in flat_input]
    return nest.pack_sequence_as(structure=inputs, flat_sequence=flat_input)


def get_batch_size(tensor):
    """Returns a unit `Tensor` representing the batch size, i.e.,
    the size of the 1st dimension of :attr:`tensor`.
    """
    return tf.shape(tensor)[0]


def get_rank(tensor):
    """Returns the tensor rank as a python `int`. The input tensor can also be
    a python array.

    Args:
        tensor: A Tensor or python array.

    Returns:
        A python `int` representing the rank of :attr:`tensor`. Returns
        `None` if the rank cannot be determined.
    """
    if tf.contrib.framework.is_tensor(tensor):
        shape = tensor.shape
        try:
            rank = len(shape.as_list())
        except ValueError: # when `shape==TensorShape(None)`
            rank = None
    else:
        array = np.asarray(tensor)
        rank = array.ndim
    return rank


def mask_sequences(sequence,
                   sequence_length,
                   dtype=None,
                   time_major=False,
                   tensor_rank=2):
    """Masks out sequence entries that are beyond the respective sequence
    lengths. Masks along the time dimension.

    :attr:`sequence` and :attr:`sequence_length` can either be python
    arrays or Tensors, respectively. If both are python arrays (or None), the
    return will be a python array as well.

    Args:
        sequence: A Tensor or python array of sequence values.
            If `time_major==False` (default), this must be a Tensor of shape
            `[batch_size, max_time, ...]`. The batch and time dimension is
            exchanged if `time_major==True`.
        sequence_length: A Tensor or python array of shape `[batch_size]`.
            Time steps beyond the respective sequence lengths will be
            made zero.
        dtype (dtype): Type of :attr:`sequence`. If `None`, infer from
            :attr:`sequence` automatically.
        time_major (bool): The shape format of the inputs. If `True`,
            :attr:`sequence` must have shape
            `[max_time, batch_size, ...]`.
            If `False` (default), :attr:`sequence` must have
            shape `[batch_size, max_time, ...]`.
        tensor_rank (int): The number of dimensions of :attr:`sequence`.
            Default is 2, i.e., :attr:`sequence` is a 2D Tensor consisting
            of batch and time dimensions. Ignored if both :attr:`sequence`
            and :attr:`sequence_length` are python arrays.

    Returns:
        The masked sequence, i.e., a Tensor or python array of the same shape
        as :attr:`sequence` but with masked-out entries (set to zero).

        If both :attr:`sequence` and :attr:`sequence_length` are python
        arrays, the returned value is a python array as well.
    """
    is_tensor = tf.contrib.framework.is_tensor
    if is_tensor(sequence) or is_tensor(sequence_length):
        return _mask_sequences_tensor(
            sequence, sequence_length, dtype, time_major, tensor_rank)
    else:
        return _mask_sequences_py(
            sequence, sequence_length, dtype, time_major)


def _mask_sequences_tensor(sequence,
                           sequence_length,
                           dtype=None,
                           time_major=False,
                           tensor_rank=2):
    """Masks out sequence entries that are beyond the respective sequence
    lengths. Masks along the time dimension.

    Args:
        sequence: A Tensor of sequence values.

            If `time_major=False` (default), this must be a Tensor of shape:
                `[batch_size, max_time, d_2, ..., d_rank]`, where the rank of
                the Tensor is specified with :attr:`tensor_rank`.

            If `time_major=True`, this must be a Tensor of shape:
                `[max_time, batch_size, d_2, ..., d_rank].`
        sequence_length: A Tensor of shape `[batch_size]`. Time steps beyond
            the respective sequence lengths will be made zero.
        dtype (dtype): Type of :attr:`sequence`. If `None`, infer from
            :attr:`sequence` automatically.
        time_major (bool): The shape format of the inputs. If `True`,
            :attr:`sequence` must have shape
            `[max_time, batch_size, d_2, ..., d_rank]`.
            If `False` (default), :attr:`sequence` must have
            shape `[batch_size, max_time, d_2, ..., d_rank]`.
        tensor_rank (int): The number of dimensions of :attr:`sequence`.
            Default is 2, i.e., :attr:`sequence` is a 2D Tensor consisting
            of batch and time dimensions.

    Returns:
        The masked sequence, i.e., a Tensor of the same shape as
        :attr:`sequence` but with masked-out entries (set to zero).
    """
    if tensor_rank is None:
        tensor_rank = 2
    if tensor_rank < 2:
        raise ValueError(
            "tensor_rank must be > 2. Got tensor_rank = {}".format(tensor_rank))
    if time_major:
        sequence = rnn._transpose_batch_time(sequence)
    max_time = tf.cast(tf.shape(sequence)[1], tf.int32)
    if dtype is None:
        dtype = sequence.dtype
    mask = tf.sequence_mask(
        tf.cast(sequence_length, tf.int32), max_time, dtype=dtype)
    for _ in range(2, tensor_rank):
        mask = tf.expand_dims(mask, axis=-1)
    sequence = sequence * mask
    if time_major:
        sequence = rnn._transpose_batch_time(sequence)
    return sequence


def _mask_sequences_py(sequence,
                       sequence_length,
                       dtype=None,
                       time_major=False):
    """Masks out sequence entries that are beyond the respective sequence
    lengths. Masks along the time dimension.

    This is the numpy version of :func:`texar.utils.mask_sequences`.

    Args:
        sequence: An python array of sequence values.

            If `time_major=False` (default), this must be an array of shape:
                `[batch_size, max_time, ...]`

            If `time_major=True`, this must be a Tensor of shape:
                `[max_time, batch_size, ...].`
        sequence_length: An array of shape `[batch_size]`. Time steps beyond
            the respective sequence lengths will be made zero.
        dtype (dtype): Type of :attr:`sequence`. If `None`, infer from
            :attr:`sequence` automatically.
        time_major (bool): The shape format of the inputs. If `True`,
            :attr:`sequence` must have shape
            `[max_time, batch_size, ...]`.
            If `False` (default), :attr:`sequence` must have
            shape `[batch_size, max_time, ...]`.

    Returns:
        The masked sequence, i.e., an array of the same shape as
        :attr:`sequence` but with masked-out entries (set to zero).
    """
    sequence = np.array(sequence)
    sequence_length = np.array(sequence_length)

    rank = sequence.ndim
    if rank < 2:
        raise ValueError("`sequence` must be 2D or higher order.")
    batch_size = sequence.shape[0]
    max_time = sequence.shape[1]
    dtype = dtype or sequence.dtype

    if time_major:
        sequence = np.transpose(sequence, axes=[1, 0, 2])

    steps = np.tile(np.arange(max_time), [batch_size, 1])
    mask = np.asarray(steps < sequence_length[:, None], dtype=dtype)
    for _ in range(2, rank):
        mask = np.expand_dims(mask, -1)

    sequence = sequence * mask

    if time_major:
        sequence = np.transpose(sequence, axes=[1, 0, 2])

    return sequence


def flatten(tensor, preserve_dims, flattened_dim=None):
    """Flattens a tensor whiling keeping several leading dimensions.

    :attr:`preserve_dims` must < tensor's rank

    Args:
        tensor: A Tensor to flatten.
        preserve_dims (int): The number of leading dimensions to preserve.
        flatterned_dim (int, optional): The size of the resulting flattened
            dimension. If not given, infer automatically, which can cause
            a statically unknown dimension size.

    Returns:
        A Tensor with rank :attr:`perserve_dims`+1.

    Example:
        .. code-block:: python

            x = tf.ones(shape=[d_1, d_2, d_3, d_4])
            y = flatten(x, 2) # y.shape == [d_1, d_2, d_3 * d_4]
    """
    if flattened_dim is None:
        flattened_dim = -1
    shape = tf.concat([tf.shape(tensor)[:preserve_dims], [flattened_dim]],
                      axis=0)
    tensor_ = tf.reshape(tensor, shape)
    return tensor_


def shape_list(x):
    """Returns **static** shape of the input Tensor whenever possible.

    Args:
        x: A Tensor.

    Returns:
        - If the rank of :attr:`x` is unknown, returns the dynamic shape: \
        `tf.shape(x)`
        - Otherwise, returns a list of dims, each of which is either an `int` \
        whenever it can be statically determined, or a scalar Tensor otherwise.
    """
    x = tf.convert_to_tensor(x)
    # If unknown rank, return dynamic shape
    if x.get_shape().dims is None:
        return tf.shape(x)
    static = x.get_shape().as_list()
    shape = tf.shape(x)
    ret = []
    for i, dim in enumerate(static):
        if dim is None:
            dim = shape[i]
        ret.append(dim)
    return ret


def pad_and_concat(values, axis, rank=None, pad_axis=None,
                   pad_constant_values=0):
    """Concats tensors along one dimension. Pads each of other dimensions of
    the tensors to the corresponding maximum size if necessary.

    Args:
        values: A list of Tensors of the same rank.
        axis (int): A Python int. Dimension along which to concatenate.
        rank (int, optional): Rank of the tensors. If `None`, inferred
            automatically from :attr:`values`.
        pad_axis (int or list, optional): A Python int or a list of int.
            Dimensions to pad. Paddings are only added to the end of
            corresponding dimensions. If `None`, all dimensions except the
            :attr:`axis` dimension are padded.
        pad_constant_values: The scalar pad value to use. Must be same type
            as the tensors.

    Returns:
        A `Tensor` resulting from padding and concatenation of the input
        tensors.

    Raises:
        ValueError: If :attr:`rank` is `None` and cannot be inferred from
            :attr:`values`.


    Example:

        .. code-block:: python

            a = tf.ones([1, 2])
            b = tf.ones([2, 3])

            c = pad_and_concat([a,b], 0)
            # c.shape == [3, 3]
            # c == [[1, 1, 0],
            #       [1, 1, 1],
            #       [1, 1, 1]]

            d = pad_and_concat([a,b], 1)
            # d.shape == [2, 5]
            # d == [[1, 1, 1, 1, 1]
            #       [0, 0, 1, 1, 1]]
    """
    if rank is None:
        for value in values:
            rank = get_rank(value)
            if rank is not None:
                break
    if rank is None:
        raise ValueError('Cannot determine the rank of the tensors')

    def _pad_to_size(value, axis_, size):
        """Pads the :attr:`axis_` of a tensor :attr:`value` to the given
        :attr:`size`. Only pads to the end.

        Args:
            value: A Tensor.
            axis_: A Python int.
            size: A scalar int Tensor or Python int.
        """
        paddings = np.zeros([rank, 2], dtype=np.int32)
        paddings[axis_, 1] = 1
        paddings = paddings * (size - tf.shape(value)[axis_])
        return tf.pad(value, paddings, mode='CONSTANT',
                      constant_values=pad_constant_values)

    if pad_axis is None:
        pad_axis = [r for r in range(rank) if r != axis]

    pad_axis = pad_axis if isinstance(pad_axis, (list, tuple)) else [pad_axis]

    for pa in pad_axis:
        max_dim_size = tf.reduce_max([tf.shape(v)[pa] for v in values])
        for i, v in enumerate(values):
            values[i] = _pad_to_size(v, pa, max_dim_size)

    return tf.concat(values, axis)



#
"""
Unit tests for conv encoders.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import tensorflow as tf

import texar as tx
from texar.modules.classifiers.conv_classifiers import Conv1DClassifier


class Conv1DClassifierTest(tf.test.TestCase):
    """Tests :class:`~texar.modules.Conv1DClassifier` class.
    """

    def test_classifier(self):
        """Tests classification.
        """
        # case 1: default hparams
        classifier = Conv1DClassifier()
        self.assertEqual(len(classifier.layers), 5)
        self.assertTrue(isinstance(classifier.layers[-1],
                                   tf.layers.Dense))
        inputs = tf.ones([64, 16, 300], tf.float32)
        logits, pred = classifier(inputs)
        self.assertEqual(logits.shape, [64, 2])
        self.assertEqual(pred.shape, [64])

        inputs = tf.placeholder(tf.float32, [64, None, 300])
        logits, pred = classifier(inputs)
        self.assertEqual(logits.shape, [64, 2])
        self.assertEqual(pred.shape, [64])

        # case 1
        hparams = {
            "num_classes": 10,
            "logit_layer_kwargs": {"use_bias": False}
        }
        classifier = Conv1DClassifier(hparams=hparams)
        inputs = tf.ones([64, 16, 300], tf.float32)
        logits, pred = classifier(inputs)
        self.assertEqual(logits.shape, [64, 10])
        self.assertEqual(pred.shape, [64])


if __name__ == "__main__":
    tf.test.main()



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Some embed_fn s used in :class:`~texar.modules.memory.MemNetBase` and its
subclasses.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=invalid-name, too-many-arguments

__all__ = [
    'default_memnet_embed_fn_hparams',
]

def default_memnet_embed_fn_hparams():
    """Returns a dictionary of hyperparameters with default hparams for
    :func:`~texar.modules.memory.default_embed_fn`

    .. code-block:: python

        {
            "embedding": {
                "dim": 100
            },
            "temporal_embedding": {
                "dim": 100
            },
            "combine_mode": "add"
        }

    Here:

    "embedding" : dict, optional
        Hyperparameters for embedding operations. See
        :meth:`~texar.modules.WordEmbedder.default_hparams` of
        :class:`~texar.modules.WordEmbedder` for details. If `None`, the
        default hyperparameters are used.

    "temporal_embedding" : dict, optional
        Hyperparameters for temporal embedding operations. See
        :meth:`~texar.modules.PositionEmbedder.default_hparams` of
        :class:`~texar.modules.PositionEmbedder` for details. If `None`, the
        default hyperparameters are used.

    "combine_mode" : str
        Either **'add'** or **'concat'**. If 'add', memory
        embedding and temporal embedding are added up. In this case the two
        embedders must have the same dimension. If 'concat', the two
        embeddings are concated.
    """
    return {
        "embedding": {
            "dim": 100
        },
        "temporal_embedding": {
            "dim": 100
        },
        "combine_mode": "add"
    }




#
"""
Unit tests for various layers.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

import tensorflow as tf
import tensorflow.contrib.rnn as rnn

import texar as tx
from texar import context
from texar.hyperparams import HParams
from texar.core import layers

# pylint: disable=no-member, protected-access, invalid-name
# pylint: disable=redefined-variable-type

class GetRNNCellTest(tf.test.TestCase):
    """Tests RNN cell creator.
    """

    def test_get_rnn_cell(self):
        """Tests :func:`texar.core.layers.get_rnn_cell`.
        """
        emb_dim = 4
        num_units = 64

        # Given instance
        hparams = {
            "type": rnn.LSTMCell(num_units)
        }
        cell = layers.get_rnn_cell(hparams)
        self.assertTrue(isinstance(cell, rnn.LSTMCell))

        # Given class
        hparams = {
            "type": rnn.LSTMCell,
            "kwargs": {"num_units": 10}
        }
        cell = layers.get_rnn_cell(hparams)
        self.assertTrue(isinstance(cell, rnn.LSTMCell))

        # Given string, and complex hyperparameters
        keep_prob_x = tf.placeholder(
            name='keep_prob', shape=[], dtype=tf.float32)
        hparams = {
            "type": "tensorflow.contrib.rnn.GRUCell",
            "kwargs": {
                "num_units": num_units
            },
            "num_layers": 2,
            "dropout": {
                "input_keep_prob": 0.8,
                "state_keep_prob": keep_prob_x,
                "variational_recurrent": True,
                "input_size": [emb_dim, num_units]
            },
            "residual": True,
            "highway": True
        }

        hparams_ = HParams(hparams, layers.default_rnn_cell_hparams())
        cell = layers.get_rnn_cell(hparams_)

        batch_size = 16
        inputs = tf.zeros([batch_size, emb_dim], dtype=tf.float32)
        output, state = cell(inputs,
                             cell.zero_state(batch_size, dtype=tf.float32))
        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())

            feed_dict = {
                keep_prob_x: 1.0,
                context.global_mode(): tf.estimator.ModeKeys.TRAIN
            }
            output_, state_ = sess.run([output, state], feed_dict=feed_dict)

            self.assertEqual(output_.shape[0], batch_size)
            if isinstance(state_, (list, tuple)):
                self.assertEqual(state_[0].shape[0], batch_size)
                self.assertEqual(state_[0].shape[1],
                                 hparams_.kwargs.num_units)
            else:
                self.assertEqual(state_.shape[0], batch_size)
                self.assertEqual(state_.shape[1],
                                 hparams_.kwargs.num_units)


    def test_switch_dropout(self):
        """Tests dropout mode.
        """
        emb_dim = 4
        num_units = 64
        hparams = {
            "kwargs": {
                "num_units": num_units
            },
            "num_layers": 2,
            "dropout": {
                "input_keep_prob": 0.8,
            },
        }
        mode = tf.placeholder(tf.string)
        hparams_ = HParams(hparams, layers.default_rnn_cell_hparams())
        cell = layers.get_rnn_cell(hparams_, mode)

        batch_size = 16
        inputs = tf.zeros([batch_size, emb_dim], dtype=tf.float32)
        output, state = cell(inputs,
                             cell.zero_state(batch_size, dtype=tf.float32))
        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            output_train, _ = sess.run(
                [output, state],
                feed_dict={mode: tf.estimator.ModeKeys.TRAIN})
            self.assertEqual(output_train.shape[0], batch_size)
            output_test, _ = sess.run(
                [output, state],
                feed_dict={mode: tf.estimator.ModeKeys.EVAL})
            self.assertEqual(output_test.shape[0], batch_size)


class GetActivationFnTest(tf.test.TestCase):
    """Tests :func:`texar.core.layers.get_activation_fn`.
    """
    def test_get_activation_fn(self):
        """Tests.
        """
        fn = layers.get_activation_fn()
        self.assertEqual(fn, tf.identity)

        fn = layers.get_activation_fn('relu')
        self.assertEqual(fn, tf.nn.relu)

        inputs = tf.random_uniform([64, 100], -5, 20, dtype=tf.int32)

        fn = layers.get_activation_fn('leaky_relu')
        fn_output = fn(inputs)
        ref_output = tf.nn.leaky_relu(inputs)
        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            fn_output_, ref_output_ = sess.run([fn_output, ref_output])
            np.testing.assert_array_equal(fn_output_, ref_output_)

        fn = layers.get_activation_fn('leaky_relu', kwargs={'alpha': 0.1})
        fn_output = fn(inputs)
        ref_output = tf.nn.leaky_relu(inputs, alpha=0.1)
        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            fn_output_, ref_output_ = sess.run([fn_output, ref_output])
            np.testing.assert_array_equal(fn_output_, ref_output_)


class GetLayerTest(tf.test.TestCase):
    """Tests layer creator.
    """
    def test_get_layer(self):
        """Tests :func:`texar.core.layers.get_layer`.
        """
        hparams = {
            "type": "Conv1D"
        }
        layer = layers.get_layer(hparams)
        self.assertTrue(isinstance(layer, tf.layers.Conv1D))

        hparams = {
            "type": "MergeLayer",
            "kwargs": {
                "layers": [
                    {"type": "Conv1D"},
                    {"type": "Conv1D"}
                ]
            }
        }
        layer = layers.get_layer(hparams)
        self.assertTrue(isinstance(layer, tx.core.MergeLayer))

        hparams = {
            "type": tf.layers.Conv1D
        }
        layer = layers.get_layer(hparams)
        self.assertTrue(isinstance(layer, tf.layers.Conv1D))

        hparams = {
            "type": tf.layers.Conv1D(filters=10, kernel_size=2)
        }
        layer = layers.get_layer(hparams)
        self.assertTrue(isinstance(layer, tf.layers.Conv1D))


class ReducePoolingLayerTest(tf.test.TestCase):
    """Tests reduce pooling layer.
    """
    def setUp(self):
        tf.test.TestCase.setUp(self)

        self._batch_size = 64
        self._seq_length = 16
        self._emb_dim = 100

    def test_max_reduce_pooling_layer(self):
        """Tests :class:`texar.core.MaxReducePooling1D`.
        """
        pool_layer = layers.MaxReducePooling1D()

        inputs = tf.random_uniform(
            [self._batch_size, self._seq_length, self._emb_dim])
        output_shape = pool_layer.compute_output_shape(inputs.get_shape())
        output = pool_layer(inputs)
        output_reduce = tf.reduce_max(inputs, axis=1)
        self.assertEqual(output.get_shape(), output_shape)
        self.assertEqual(output.get_shape(), [self._batch_size, self._emb_dim])

        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            output_, output_reduce_ = sess.run([output, output_reduce])
            np.testing.assert_array_equal(output_, output_reduce_)

    def test_average_reduce_pooling_layer(self):
        """Tests :class:`texar.core.AverageReducePooling1D`.
        """
        pool_layer = layers.AverageReducePooling1D()

        inputs = tf.random_uniform(
            [self._batch_size, self._seq_length, self._emb_dim])
        output_shape = pool_layer.compute_output_shape(inputs.get_shape())
        output = pool_layer(inputs)
        output_reduce = tf.reduce_mean(inputs, axis=1)
        self.assertEqual(output.get_shape(), output_shape)
        self.assertEqual(output.get_shape(), [self._batch_size, self._emb_dim])

        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            output_, output_reduce_ = sess.run([output, output_reduce])
            np.testing.assert_array_equal(output_, output_reduce_)

class MergeLayerTest(tf.test.TestCase):
    """Tests MergeLayer.
    """

    def test_output_shape(self):
        """Tests MergeLayer.compute_output_shape function.
        """
        input_shapes = [[None, 1, 2], [64, 2, 2], [None, 3, 2]]

        concat_layer = layers.MergeLayer(mode='concat', axis=1)
        concat_output_shape = concat_layer.compute_output_shape(input_shapes)
        self.assertEqual(concat_output_shape, [64, 6, 2])

        sum_layer = layers.MergeLayer(mode='sum', axis=1)
        sum_output_shape = sum_layer.compute_output_shape(input_shapes)
        self.assertEqual(sum_output_shape, [64, 2])

        input_shapes = [[None, 5, 2], [64, None, 2], [2]]
        esum_layer = layers.MergeLayer(mode='elemwise_sum')
        esum_output_shape = esum_layer.compute_output_shape(input_shapes)
        self.assertEqual(esum_output_shape, [64, 5, 2])

    def test_layer_logics(self):
        """Test the logic of MergeLayer.
        """
        layers_ = []
        layers_.append(tf.layers.Conv1D(filters=200, kernel_size=3))
        layers_.append(tf.layers.Conv1D(filters=200, kernel_size=4))
        layers_.append(tf.layers.Conv1D(filters=200, kernel_size=5))
        layers_.append(tf.layers.Dense(200))
        layers_.append(tf.layers.Dense(200))
        m_layer = layers.MergeLayer(layers_)

        inputs = tf.zeros([64, 16, 1024], dtype=tf.float32)
        outputs = m_layer(inputs)
        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            outputs_ = sess.run(outputs)
            self.assertEqual(outputs_.shape[0], 64)
            self.assertEqual(outputs_.shape[2], 200)
            self.assertEqual(
                outputs_.shape,
                m_layer.compute_output_shape(inputs.shape.as_list()))

    def test_trainable_variables(self):
        """Test the trainable_variables of the layer.
        """
        layers_ = []
        layers_.append(tf.layers.Conv1D(filters=200, kernel_size=3))
        layers_.append(tf.layers.Conv1D(filters=200, kernel_size=4))
        layers_.append(tf.layers.Conv1D(filters=200, kernel_size=5))
        layers_.append(tf.layers.Dense(200))
        layers_.append(tf.layers.Dense(200))
        m_layer = layers.MergeLayer(layers_)

        inputs = tf.zeros([64, 16, 1024], dtype=tf.float32)
        _ = m_layer(inputs)

        num_vars = sum([len(layer.trainable_variables) for layer in layers_])
        self.assertEqual(num_vars, len(m_layer.trainable_variables))

class SequentialLayerTest(tf.test.TestCase):
    """Tests sequential layer.
    """

    def test_seq_layer(self):
        """Test sequential layer.
        """
        layers_ = []
        layers_.append(tf.layers.Dense(100))
        layers_.append(tf.layers.Dense(200))
        seq_layer = layers.SequentialLayer(layers_)

        output_shape = seq_layer.compute_output_shape([None, 10])
        self.assertEqual(output_shape[1].value, 200)

        inputs = tf.zeros([10, 20], dtype=tf.float32)
        outputs = seq_layer(inputs)

        num_vars = sum([len(layer.trainable_variables) for layer in layers_])
        self.assertEqual(num_vars, len(seq_layer.trainable_variables))

        with self.test_session() as sess:
            sess.run(tf.global_variables_initializer())
            outputs_ = sess.run(outputs)
            self.assertEqual(outputs_.shape[0], 10)
            self.assertEqual(outputs_.shape[1], 200)


if __name__ == "__main__":
    tf.test.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions related to BERT encoders.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import json
import collections
import re
import sys
import os
import tensorflow as tf
from texar.data.data_utils import maybe_download
__all__ = [
    "transform_bert_to_texar_config",
    "init_bert_checkpoint",
    "load_pretrained_model"
]

_BERT_PATH = "https://storage.googleapis.com/bert_models/"
_MODEL2URL = {
    'bert-base-uncased':
        _BERT_PATH + "2018_10_18/uncased_L-12_H-768_A-12.zip",
    'bert-large-uncased':
        _BERT_PATH + "2018_10_18/uncased_L-24_H-1024_A-16.zip",
    'bert-base-cased':
        _BERT_PATH + "2018_10_18/cased_L-12_H-768_A-12.zip",
    'bert-large-cased':
        _BERT_PATH + "2018_10_18/cased_L-24_H-1024_A-16.zip",
    'bert-base-multilingual-uncased':
        _BERT_PATH + "2018_11_23/multi_cased_L-12_H-768_A-12.zip",
    'bert-base-multilingual-cased':
        _BERT_PATH + "2018_11_03/multilingual_L-12_H-768_A-12.zip",
    'bert-base-chinese':
        _BERT_PATH + "2018_11_03/chinese_L-12_H-768_A-12.zip",
}

def _get_assignment_map_from_checkpoint(tvars, init_checkpoint, scope_name):
    """
    Provided by Google AI Language Team.
    Compute the union of the current variables and checkpoint variables.
    Because the variable scope of the original BERT and Texar implementation,
    we need to build a assignment map to match the variables.
    """
    initialized_variable_names = {}

    name_to_variable = collections.OrderedDict()
    for var in tvars:
        name = var.name
        m = re.match("^(.*):\\d+$", name)
        if m is not None:
            name = m.group(1)
        name_to_variable[name] = var

    init_vars = tf.train.list_variables(init_checkpoint)

    assignment_map = {
        'bert/embeddings/word_embeddings':
            scope_name + '/word_embeddings/w',
        'bert/embeddings/token_type_embeddings':
            scope_name + '/token_type_embeddings/w',
        'bert/embeddings/position_embeddings':
            scope_name + '/position_embeddings/w',
        'bert/embeddings/LayerNorm/beta':
            scope_name + '/encoder/LayerNorm/beta',
        'bert/embeddings/LayerNorm/gamma':
            scope_name + '/encoder/LayerNorm/gamma',
    }
    for check_name, model_name in assignment_map.items():
        initialized_variable_names[model_name] = 1
        initialized_variable_names[model_name + ":0"] = 1

    for check_name, _ in init_vars:
        if check_name.startswith('bert'):
            if check_name.startswith('bert/embeddings'):
                continue
            check_name_scope = check_name.replace("bert/", scope_name+'/')
            model_name = re.sub(
                'layer_\\d+/output/dense',
                lambda x: x.group(0).replace('output/dense', 'ffn/output'),
                check_name_scope)
            if model_name == check_name_scope:
                model_name = re.sub(
                    'layer_\\d+/output/LayerNorm',
                    lambda x: x.group(0).replace('output/LayerNorm',
                                                 'ffn/LayerNorm'),
                    check_name_scope)
            if model_name == check_name_scope:
                model_name = re.sub(
                    'layer_\\d+/intermediate/dense',
                    lambda x: x.group(0).replace('intermediate/dense',
                                                 'ffn/intermediate'),
                    check_name_scope)
            if model_name == check_name_scope:
                model_name = re.sub('attention/output/dense',
                                    'attention/self/output', check_name_scope)
            if model_name == check_name_scope:
                model_name = check_name_scope.replace(
                    'attention/output/LayerNorm', 'output/LayerNorm')

            if model_name in name_to_variable.keys():
                assignment_map[check_name] = model_name
                initialized_variable_names[model_name] = 1
                initialized_variable_names[model_name + ":0"] = 1
            else:
                tf.logging.info('model name:{} not exist'.format(model_name))

    return assignment_map, initialized_variable_names


def init_bert_checkpoint(init_checkpoint_dir, scope_name):
    """
    Initializes BERT model parameters from a checkpoint.
    Provided by Google AI Language Team.

    Args:
        init_checkpoint_dir (str): path to the checkpoint.
        scope_name: variable scope of bert encoder.
    """
    tvars = tf.trainable_variables()
    init_checkpoint = os.path.join(init_checkpoint_dir, 'bert_model.ckpt')
    if init_checkpoint:
        (assignment_map, initialized_variable_names
        ) = _get_assignment_map_from_checkpoint(
            tvars, init_checkpoint, scope_name)
        tf.train.init_from_checkpoint(init_checkpoint, assignment_map)


def _default_download_dir():
    """
    Return the directory to which packages will be downloaded by default.
    """
    package_dir = os.path.dirname(os.path.dirname(
        os.path.dirname(os.path.dirname(__file__))))
    if os.access(package_dir, os.W_OK):
        texar_download_dir = os.path.join(package_dir, 'texar_download')
    else:
        # On Windows, use %APPDATA%
        if sys.platform == 'win32' and 'APPDATA' in os.environ:
            home_dir = os.environ['APPDATA']

        # Otherwise, install in the user's home directory.
        else:
            home_dir = os.path.expanduser('~/')
            if home_dir == '~/':
                raise ValueError("Could not find a default download directory")

        texar_download_dir = os.path.join(home_dir, 'texar_download')

    if not os.path.exists(texar_download_dir):
        os.mkdir(texar_download_dir)

    return os.path.join(texar_download_dir, 'bert')


def load_pretrained_model(pretrained_model_name, cache_dir):
    """
    Return the directory in which the pretrained model is cached.
    """
    if pretrained_model_name in _MODEL2URL:
        download_path = _MODEL2URL[pretrained_model_name]
    else:
        raise ValueError(
            "Pre-trained model not found: {}".format(pretrained_model_name))

    if cache_dir is None:
        cache_dir = _default_download_dir()

    file_name = download_path.split('/')[-1]

    cache_path = os.path.join(cache_dir, file_name.split('.')[0])
    if not os.path.exists(cache_path):
        maybe_download(download_path, cache_dir, extract=True)
    else:
        print("Using cached pre-trained BERT model from: %s." % cache_path)

    return cache_path


def transform_bert_to_texar_config(config_dir):
    """
    Load the Json config file and transform it into Texar style configuration.
    """
    config_ckpt = json.loads(
        open(os.path.join(config_dir, 'bert_config.json')).read())
    configs = {}
    hidden_dim = config_ckpt['hidden_size']
    configs['hidden_size'] = config_ckpt['hidden_size']
    configs['embed'] = {
        'name': 'word_embeddings',
        'dim': hidden_dim}
    configs['vocab_size'] = config_ckpt['vocab_size']

    configs['segment_embed'] = {
        'name': 'token_type_embeddings',
        'dim': hidden_dim}
    configs['type_vocab_size'] = config_ckpt['type_vocab_size']

    configs['position_embed'] = {
        'name': 'position_embeddings',
        'dim': hidden_dim}
    configs['position_size'] = config_ckpt['max_position_embeddings']

    configs['encoder'] = {
        'name': 'encoder',
        'embedding_dropout': config_ckpt['hidden_dropout_prob'],
        'num_blocks': config_ckpt['num_hidden_layers'],
        'multihead_attention': {
            'use_bias': True,
            'num_units': hidden_dim,
            'num_heads': config_ckpt['num_attention_heads'],
            'output_dim': hidden_dim,
            'dropout_rate': config_ckpt['attention_probs_dropout_prob'],
            'name': 'self'
        },
        'residual_dropout': config_ckpt['hidden_dropout_prob'],
        'dim': hidden_dim,
        'use_bert_config': True,
        'poswise_feedforward': {
            "layers": [
                {
                    'type': 'Dense',
                    'kwargs': {
                        'name': 'intermediate',
                        'units': config_ckpt['intermediate_size'],
                        'activation': config_ckpt['hidden_act'],
                        'use_bias': True,
                    }
                },
                {
                    'type': 'Dense',
                    'kwargs': {
                        'name': 'output',
                        'units': hidden_dim,
                        'activation': None,
                        'use_bias': True,
                    }
                },
            ],
        },
    }
    return configs



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various data iterator classes.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import tensorflow as tf

import texar as tx
from texar.utils.variables import get_unique_named_variable_scope

__all__ = [
    "DataIteratorBase",
    "DataIterator",
    "TrainTestDataIterator",
    "FeedableDataIterator",
    "TrainTestFeedableDataIterator"
]

class DataIteratorBase(object):
    """Base class for all data iterator classes to inherit. A data iterator
    is a wrapper of :tf_main:`tf.data.Iterator <data/Iterator>`, and can
    switch between and iterate through **multiple** datasets.

    Args:
        datasets: Datasets to iterates through. This can be:

            - A single instance of :tf_main:`tf.data.Dataset <data/Dataset>` \
            or instance of subclass of :class:`~texar.data.DataBase`.
            - A `dict` that maps dataset name to \
            instance of :tf_main:`tf.data.Dataset <data/Dataset>` or \
            subclass of :class:`~texar.data.DataBase`.
            - A `list` of instances of subclasses of \
            :class:`texar.data.DataBase`. The name of instances \
            (:attr:`texar.data.DataBase.name`) must be unique.
    """

    def __init__(self, datasets):
        self._default_dataset_name = 'data'
        if isinstance(datasets, (tf.data.Dataset, tx.data.DataBase)):
            datasets = {self._default_dataset_name: datasets}
        elif isinstance(datasets, (list, tuple)):
            if any(not isinstance(d, tx.data.DataBase) for d in datasets):
                raise ValueError("`datasets` must be an non-empty list of "
                                 "`texar.data.DataBase` instances.")
            num_datasets = len(datasets)
            datasets = {d.name: d for d in datasets}
            if len(datasets) < num_datasets:
                raise ValueError("Names of datasets must be unique.")

        _datasets = {}
        for k, v in datasets.items(): # pylint: disable=invalid-name
            _datasets[k] = v if isinstance(v, tf.data.Dataset) else v.dataset
        self._datasets = _datasets

        if len(self._datasets) <= 0:
            raise ValueError("`datasets` must not be empty.")

    @property
    def num_datasets(self):
        """Number of datasets.
        """
        return len(self._datasets)

    @property
    def dataset_names(self):
        """A list of dataset names.
        """
        return list(self._datasets.keys())


class DataIterator(DataIteratorBase):
    """Data iterator that switches and iterates through multiple datasets.

    This is a wrapper of TF reinitializble :tf_main:`iterator <data/Iterator>`.

    Args:
        datasets: Datasets to iterates through. This can be:

            - A single instance of :tf_main:`tf.data.Dataset <data/Dataset>` \
            or instance of subclass of :class:`~texar.data.DataBase`.
            - A `dict` that maps dataset name to \
            instance of :tf_main:`tf.data.Dataset <data/Dataset>` or \
            subclass of :class:`~texar.data.DataBase`.
            - A `list` of instances of subclasses of \
            :class:`texar.data.DataBase`. The name of instances \
            (:attr:`texar.data.DataBase.name`) must be unique.

    Example:

        .. code-block:: python

            train_data = MonoTextData(hparams_train)
            test_data = MonoTextData(hparams_test)
            iterator = DataIterator({'train': train_data, 'test': test_data})
            batch = iterator.get_next()

            sess = tf.Session()

            for _ in range(200): # Run 200 epochs of train/test
                # Starts iterating through training data from the beginning
                iterator.switch_to_dataset(sess, 'train')
                while True:
                    try:
                        train_batch_ = sess.run(batch)
                    except tf.errors.OutOfRangeError:
                        print("End of training epoch.")
                # Starts iterating through test data from the beginning
                iterator.switch_to_dataset(sess, 'test')
                while True:
                    try:
                        test_batch_ = sess.run(batch)
                    except tf.errors.OutOfRangeError:
                        print("End of test epoch.")
    """

    def __init__(self, datasets):
        DataIteratorBase.__init__(self, datasets)

        self._variable_scope = get_unique_named_variable_scope('data_iterator')
        with tf.variable_scope(self._variable_scope):
            first_dataset = self._datasets[sorted(self.dataset_names)[0]]
            self._iterator = tf.data.Iterator.from_structure(
                first_dataset.output_types, first_dataset.output_shapes)
            self._iterator_init_ops = {
                name: self._iterator.make_initializer(d)
                for name, d in self._datasets.items()
            }

    def switch_to_dataset(self, sess, dataset_name=None):
        """Re-initializes the iterator of a given dataset and starts iterating
        over the dataset (from the beginning).

        Args:
            sess: The current tf session.
            dataset_name (optional): Name of the dataset. If not provided,
                there must be only one Dataset.
        """
        if dataset_name is None:
            if self.num_datasets > 1:
                raise ValueError("`dataset_name` is required if there are "
                                 "more than one datasets.")
            dataset_name = next(iter(self._datasets))
        if dataset_name not in self._datasets:
            raise ValueError("Dataset not found: ", dataset_name)
        sess.run(self._iterator_init_ops[dataset_name])

    def get_next(self):
        """Returns the next element of the activated dataset.
        """
        return self._iterator.get_next()

class TrainTestDataIterator(DataIterator):
    """Data iterator that alternatives between train, val, and test datasets.

    :attr:`train`, :attr:`val`, and :attr:`test` can be instance of
    either :tf_main:`tf.data.Dataset <data/Dataset>` or subclass of
    :class:`~texar.data.DataBase`. At least one of them must be provided.

    This is a wrapper of :class:`~texar.data.DataIterator`.

    Args:
        train (optional): Training data.
        val (optional): Validation data.
        test (optional): Test data.

    Example:

        .. code-block:: python

            train_data = MonoTextData(hparams_train)
            val_data = MonoTextData(hparams_val)
            iterator = TrainTestDataIterator(train=train_data, val=val_data)
            batch = iterator.get_next()

            sess = tf.Session()

            for _ in range(200): # Run 200 epochs of train/val
                # Starts iterating through training data from the beginning
                iterator.switch_to_train_data(sess)
                while True:
                    try:
                        train_batch_ = sess.run(batch)
                    except tf.errors.OutOfRangeError:
                        print("End of training epoch.")
                # Starts iterating through val data from the beginning
                iterator.switch_to_val_dataset(sess)
                while True:
                    try:
                        val_batch_ = sess.run(batch)
                    except tf.errors.OutOfRangeError:
                        print("End of val epoch.")
    """

    def __init__(self, train=None, val=None, test=None):
        dataset_dict = {}
        self._train_name = 'train'
        self._val_name = 'val'
        self._test_name = 'test'
        if train is not None:
            dataset_dict[self._train_name] = train
        if val is not None:
            dataset_dict[self._val_name] = val
        if test is not None:
            dataset_dict[self._test_name] = test
        if len(dataset_dict) == 0:
            raise ValueError("At least one of `train`, `val`, and `test` "
                             "must be provided.")

        DataIterator.__init__(self, dataset_dict)


    def switch_to_train_data(self, sess):
        """Starts to iterate through training data (from the beginning).

        Args:
            sess: The current tf session.
        """
        if self._train_name not in self._datasets:
            raise ValueError("Training data not provided.")
        self.switch_to_dataset(sess, self._train_name)

    def switch_to_val_data(self, sess):
        """Starts to iterate through val data (from the beginning).

        Args:
            sess: The current tf session.
        """
        if self._val_name not in self._datasets:
            raise ValueError("Val data not provided.")
        self.switch_to_dataset(sess, self._val_name)

    def switch_to_test_data(self, sess):
        """Starts to iterate through test data (from the beginning).

        Args:
            sess: The current tf session.
        """
        if self._test_name not in self._datasets:
            raise ValueError("Test data not provided.")
        self.switch_to_dataset(sess, self._test_name)

class FeedableDataIterator(DataIteratorBase):
    """Data iterator that iterates through **multiple** datasets and switches
    between datasets.

    The iterator can switch to a dataset and resume from where we
    left off last time we visited the dataset. This is a wrapper of TF
    feedable :tf_main:`iterator <data/Iterator>`.

    Args:
        datasets: Datasets to iterates through. This can be:

            - A single instance of :tf_main:`tf.data.Dataset <data/Dataset>` \
            or instance of subclass of :class:`~texar.data.DataBase`.
            - A `dict` that maps dataset name to \
            instance of :tf_main:`tf.data.Dataset <data/Dataset>` or \
            subclass of :class:`~texar.data.DataBase`.
            - A `list` of instances of subclasses of \
            :class:`texar.data.DataBase`. The name of instances \
            (:attr:`texar.data.DataBase.name`) must be unique.

    Example:

        .. code-block:: python

            train_data = MonoTextData(hparams={'num_epochs': 200, ...})
            test_data = MonoTextData(hparams_test)
            iterator = FeedableDataIterator({'train': train_data,
                                             'test': test_data})
            batch = iterator.get_next()

            sess = tf.Session()

            def _eval_epoch(): # Iterate through test data for one epoch
                # Initialize and start from beginning of test data
                iterator.initialize_dataset(sess, 'test')
                while True:
                    try:
                        fetch_dict = { # Read from test data
                            iterator.handle: Iterator.get_handle(sess, 'test')
                        }
                        test_batch_ = sess.run(batch, feed_dict=feed_dict)
                    except tf.errors.OutOfRangeError:
                        print("End of val epoch.")

            # Initialize and start from beginning of training data
            iterator.initialize_dataset(sess, 'train')
            step = 0
            while True:
                try:
                    fetch_dict = { # Read from training data
                        iterator.handle: Iterator.get_handle(sess, 'train')
                    }
                    train_batch_ = sess.run(batch, fetch_dict=fetch_dict)

                    step +=1
                    if step % 200 == 0: # Evaluate periodically
                        _eval_epoch()
                except tf.errors.OutOfRangeError:
                    print("End of training.")
    """

    def __init__(self, datasets):
        DataIteratorBase.__init__(self, datasets)

        self._variable_scope = get_unique_named_variable_scope(
            'feedable_data_iterator')
        with tf.variable_scope(self._variable_scope):
            self._handle = tf.placeholder(tf.string, shape=[], name='handle')
            first_dataset = self._datasets[sorted(self.dataset_names)[0]]
            self._iterator = tf.data.Iterator.from_string_handle(
                self._handle, first_dataset.output_types,
                first_dataset.output_shapes)

            self._dataset_iterators = {
                name: dataset.make_initializable_iterator()
                for name, dataset in self._datasets.items()
            }

    def get_handle(self, sess, dataset_name=None):
        """Returns a dataset handle used to feed the
        :attr:`handle` placeholder to fetch data from the dataset.

        Args:
            sess: The current tf session.
            dataset_name (optional): Name of the dataset. If not provided,
                there must be only one Dataset.

        Returns:
            A string handle to be fed to the :attr:`handle` placeholder.

        Example:

            .. code-block:: python

                next_element = iterator.get_next()
                train_handle = iterator.get_handle(sess, 'train')
                # Gets the next training element
                ne_ = sess.run(next_element,
                               feed_dict={iterator.handle: train_handle})
        """
        if dataset_name is None:
            if self.num_datasets > 1:
                raise ValueError("`dataset_name` is required if there are "
                                 "more than one datasets.")
            dataset_name = next(iter(self._datasets))
        if dataset_name not in self._datasets:
            raise ValueError("Dataset not found: ", dataset_name)
        return sess.run(self._dataset_iterators[dataset_name].string_handle())

    def restart_dataset(self, sess, dataset_name=None):
        """Restarts datasets so that next iteration will fetch data from
        the beginning of the datasets.

        Args:
            sess: The current tf session.
            dataset_name (optional): A dataset name or a list of dataset names
                that specifies which dataset(s) to restart. If `None`, all
                datasets are restart.
        """
        self.initialize_dataset(sess, dataset_name)

    def initialize_dataset(self, sess, dataset_name=None):
        """Initializes datasets. A dataset must be initialized before being
        used.

        Args:
            sess: The current tf session.
            dataset_name (optional): A dataset name or a list of dataset names
                that specifies which dataset(s) to initialize. If `None`, all
                datasets are initialized.
        """
        if dataset_name is None:
            dataset_name = self.dataset_names
        if not isinstance(dataset_name, (tuple, list)):
            dataset_name = [dataset_name]

        for name in dataset_name:
            sess.run(self._dataset_iterators[name].initializer)

    def get_next(self):
        """Returns the next element of the activated dataset.
        """
        return self._iterator.get_next()

    @property
    def handle(self):
        """The handle placeholder that can be fed with a dataset handle to
        fetch data from the dataset.
        """
        return self._handle

class TrainTestFeedableDataIterator(FeedableDataIterator):
    """Feedable data iterator that alternatives between train, val, and test
    datasets.

    This is a wrapper of :class:`~texar.data.FeedableDataIterator`.
    The iterator can switch to a dataset and resume from where it was
    left off when it was visited last time.

    :attr:`train`, :attr:`val`, and :attr:`test` can be instance of
    either :tf_main:`tf.data.Dataset <data/Dataset>` or subclass of
    :class:`~texar.data.DataBase`. At least one of them must be provided.

    Args:
        train (optional): Training data.
        val (optional): Validation data.
        test (optional): Test data.

    Example:

        .. code-block:: python

            train_data = MonoTextData(hparams={'num_epochs': 200, ...})
            test_data = MonoTextData(hparams_test)
            iterator = TrainTestFeedableDataIterator(train=train_data,
                                                     test=test_data)
            batch = iterator.get_next()

            sess = tf.Session()

            def _eval_epoch(): # Iterate through test data for one epoch
                # Initialize and start from beginning of test data
                iterator.initialize_test_dataset(sess)
                while True:
                    try:
                        fetch_dict = { # Read from test data
                            iterator.handle: Iterator.get_test_handle(sess)
                        }
                        test_batch_ = sess.run(batch, feed_dict=feed_dict)
                    except tf.errors.OutOfRangeError:
                        print("End of test epoch.")

            # Initialize and start from beginning of training data
            iterator.initialize_train_dataset(sess)
            step = 0
            while True:
                try:
                    fetch_dict = { # Read from training data
                        iterator.handle: Iterator.get_train_handle(sess)
                    }
                    train_batch_ = sess.run(batch, fetch_dict=fetch_dict)

                    step +=1
                    if step % 200 == 0: # Evaluate periodically
                        _eval_epoch()
                except tf.errors.OutOfRangeError:
                    print("End of training.")
    """

    def __init__(self, train=None, val=None, test=None):
        dataset_dict = {}
        self._train_name = 'train'
        self._val_name = 'val'
        self._test_name = 'test'
        if train is not None:
            dataset_dict[self._train_name] = train
        if val is not None:
            dataset_dict[self._val_name] = val
        if test is not None:
            dataset_dict[self._test_name] = test
        if len(dataset_dict) == 0:
            raise ValueError("At least one of `train`, `val`, and `test` "
                             "must be provided.")

        FeedableDataIterator.__init__(self, dataset_dict)

    def get_train_handle(self, sess):
        """Returns the handle of the training dataset. The handle can be used
        to feed the :attr:`handle` placeholder to fetch training data.

        Args:
            sess: The current tf session.

        Returns:
            A string handle to be fed to the :attr:`handle` placeholder.

        Example:

            .. code-block:: python

                next_element = iterator.get_next()
                train_handle = iterator.get_train_handle(sess)
                # Gets the next training element
                ne_ = sess.run(next_element,
                               feed_dict={iterator.handle: train_handle})
        """
        if self._train_name not in self._datasets:
            raise ValueError("Training data not provided.")
        return self.get_handle(sess, self._train_name)

    def get_val_handle(self, sess):
        """Returns the handle of the validation dataset. The handle can be used
        to feed the :attr:`handle` placeholder to fetch validation data.

        Args:
            sess: The current tf session.

        Returns:
            A string handle to be fed to the :attr:`handle` placeholder.
        """
        if self._val_name not in self._datasets:
            raise ValueError("Val data not provided.")
        return self.get_handle(sess, self._val_name)

    def get_test_handle(self, sess):
        """Returns the handle of the test dataset. The handle can be used
        to feed the :attr:`handle` placeholder to fetch test data.

        Args:
            sess: The current tf session.

        Returns:
            A string handle to be fed to the :attr:`handle` placeholder.
        """
        if self._test_name not in self._datasets:
            raise ValueError("Test data not provided.")
        return self.get_handle(sess, self._test_name)

    def restart_train_dataset(self, sess):
        """Restarts the training dataset so that next iteration will fetch
        data from the beginning of the training dataset.

        Args:
            sess: The current tf session.
        """
        if self._train_name not in self._datasets:
            raise ValueError("Training data not provided.")
        self.restart_dataset(sess, self._train_name)

    def restart_val_dataset(self, sess):
        """Restarts the validation dataset so that next iteration will fetch
        data from the beginning of the validation dataset.

        Args:
            sess: The current tf session.
        """
        if self._val_name not in self._datasets:
            raise ValueError("Val data not provided.")
        self.restart_dataset(sess, self._val_name)

    def restart_test_dataset(self, sess):
        """Restarts the test dataset so that next iteration will fetch
        data from the beginning of the test dataset.

        Args:
            sess: The current tf session.
        """
        if self._test_name not in self._datasets:
            raise ValueError("Test data not provided.")
        self.restart_dataset(sess, self._test_name)



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for episodic reinforcement learning agents.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from texar.agents.agent_base import AgentBase

# pylint: disable=too-many-instance-attributes

class EpisodicAgentBase(AgentBase):
    """Base class inherited by episodic RL agents.

    An agent is a wrapper of the **training process** that trains a model
    with RL algorithms. Agent itself does not create new trainable variables.

    An episodic RL agent typically provides 3 interfaces, namely, :meth:`reset`,
    :meth:`get_action` and :meth:`observe`, and is used as the following
    example.

    Example:

        .. code-block:: python

            env = SomeEnvironment(...)
            agent = PGAgent(...)

            while True:
                # Starts one episode
                agent.reset()
                observ = env.reset()
                while True:
                    action = agent.get_action(observ)
                    next_observ, reward, terminal = env.step(action)
                    agent.observe(reward, terminal)
                    observ = next_observ
                    if terminal:
                        break

    Args:
        env_config: An instance of :class:`~texar.agents.EnvConfig` specifying
            action space, observation space, and reward range, etc. Use
            :func:`~texar.agents.get_gym_env_config` to create an EnvConfig
            from a gym environment.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparamerter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter sturcture and
            default values.
    """
    def __init__(self, env_config, hparams=None):
        AgentBase.__init__(self, hparams)

        self._env_config = env_config

        self._reset_tmplt_fn = tf.make_template(
            "{}_reset".format(self.name), self._reset)
        self._observe_tmplt_fn = tf.make_template(
            "{}_observe".format(self.name), self._observe)
        self._get_action_tmplt_fn = tf.make_template(
            "{}_get_action".format(self.name), self._get_action)

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "name": "agent"
            }
        """
        return {
            'name': 'agent'
        }

    def reset(self):
        """Resets the states to begin new episode.
        """
        self._reset_tmplt_fn()

    def _reset(self):
        raise NotImplementedError

    def observe(self, reward, terminal, train_policy=True, feed_dict=None):
        """Observes experience from environment.

        Args:
            reward: Reward of the action. The configuration (e.g., shape) of
                the reward is defined in :attr:`env_config`.
            terminal (bool): Whether the episode is terminated.
            train_policy (bool): Wether to update the policy for this step.
            feed_dict (dict, optional): Any stuffs fed to running the training
                operator.
        """
        return self._observe_tmplt_fn(reward, terminal, train_policy, feed_dict)

    def _observe(self, reward, terminal, train_policy, feed_dict):
        raise NotImplementedError

    def get_action(self, observ, feed_dict=None):
        """Gets action according to observation.

        Args:
            observ: Observation from the environment.

        Returns:
            action from the policy.
        """
        return self._get_action_tmplt_fn(observ, feed_dict)

    def _get_action(self, observ, feed_dict):
        raise NotImplementedError

    @property
    def env_config(self):
        """Environment configuration.
        """
        return self._env_config



"""
Unit tests for shape-related utility functions.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=no-member

import numpy as np

import tensorflow as tf

from texar.utils import shapes

class ShapesTest(tf.test.TestCase):
    """Tests shape-related utility functions.
    """

    def test_mask_sequences(self):
        """Tests :func:`texar.utils.shapes.mask_sequences`.
        """
        seq = np.ones([3, 4, 3], dtype=np.int32)
        seq_length = np.array([3, 2, 1], dtype=np.int32)

        masked_seq = shapes.mask_sequences(seq, seq_length)
        self.assertEqual(masked_seq.shape, seq.shape)
        seq_sum = np.sum(masked_seq, axis=(1, 2))
        np.testing.assert_array_equal(seq_sum, seq_length * 3)

    def test_pad_and_concat(self):
        """Test :func:`texar.utils.shapes.pad_and_concat`.
        """
        a = tf.ones([3, 10, 2])
        b = tf.ones([4, 20, 3])
        c = tf.ones([5, 1, 4])

        t = shapes.pad_and_concat([a, b, c], 0)
        self.assertEqual(t.shape, [3+4+5, 20, 4])
        t = shapes.pad_and_concat([a, b, c], 1)
        self.assertEqual(t.shape, [5, 10+20+1, 4])
        t = shapes.pad_and_concat([a, b, c], 2)
        self.assertEqual(t.shape, [5, 20, 2+3+4])

        d = tf.placeholder(dtype=tf.float32, shape=[6, None, 1])
        t = shapes.pad_and_concat([a, b, c, d], 0)
        with self.test_session() as sess:
            t_ = sess.run(t, feed_dict={d: np.ones([6, 2, 1])})
            self.assertEqual(list(t_.shape), [3+4+5+6, 20, 4])

if __name__ == "__main__":
    tf.test.main()




# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions related to variables.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

# pylint: disable=invalid-name

import tensorflow as tf

__all__ = [
    "get_unique_named_variable_scope",
    "add_variable",
    "collect_trainable_variables"
]


def get_unique_named_variable_scope(base_name):
    """Returns a variable scope with a unique name.

    Args:
        base_name (str): The base name to uniquified.

    Returns:
        An instance of :tf_main:`variable_scope <variable_scope>`.

    Example:

        .. code-block:: python

            vs = get_unique_named_variable_scope('base_name')
            with tf.variable_scope(vs):
                ....
    """
    with tf.variable_scope(None, default_name=base_name) as vs:
        return vs

def add_variable(variable, var_list):
    """Adds variable to a given list.

    Args:
        variable: A (list of) variable(s).
        var_list (list): The list where the :attr:`variable` are added to.
    """
    if isinstance(variable, (list, tuple)):
        for var in variable:
            add_variable(var, var_list)
    else:
        if variable not in var_list:
            var_list.append(variable)

def collect_trainable_variables(modules):
    """Collects all trainable variables of modules.

    Trainable variables included in multiple modules occur only once in the
    returned list.

    Args:
        modules: A (list of) instance of the subclasses of
            :class:`~texar.modules.ModuleBase`.

    Returns:
        A list of trainable variables in the modules.
    """
    if not isinstance(modules, (list, tuple)):
        modules = [modules]

    var_list = []
    for mod in modules:
        add_variable(mod.trainable_variables, var_list)

    return var_list



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Adversarial losses.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

def binary_adversarial_losses(real_data,
                              fake_data,
                              discriminator_fn,
                              mode="max_real"):
    """Computes adversarial losses of real/fake binary discrimination game.

    .. role:: python(code)
       :language: python

    Args:
        real_data (Tensor or array): Real data of shape
            `[num_real_examples, ...]`.
        fake_data (Tensor or array): Fake data of shape
            `[num_fake_examples, ...]`. `num_real_examples` does not
            necessarily equal `num_fake_examples`.
        discriminator_fn: A callable takes data (e.g., :attr:`real_data` and
            :attr:`fake_data`) and returns the logits of being real. The
            signature of `discriminator_fn` must be:
            :python:`logits, ... = discriminator_fn(data)`.
            The return value of `discriminator_fn` can be the logits, or
            a tuple where the logits are the first element.

        mode (str): Mode of the generator loss. Either "max_real" or "min_fake".

            - **"max_real"** (default): minimizing the generator loss is to\
            maximize the probability of fake data being classified as real.

            - **"min_fake"**: minimizing the generator loss is to minimize the\
            probability of fake data being classified as fake.

    Returns:
        A tuple `(generator_loss, discriminator_loss)` each of which is
        a scalar Tensor, loss to be minimized.
    """
    real_logits = discriminator_fn(real_data)
    if isinstance(real_logits, (list, tuple)):
        real_logits = real_logits[0]
    real_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(
        logits=real_logits, labels=tf.ones_like(real_logits)))

    fake_logits = discriminator_fn(fake_data)
    if isinstance(fake_logits, (list, tuple)):
        fake_logits = fake_logits[0]
    fake_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(
        logits=fake_logits, labels=tf.zeros_like(fake_logits)))

    d_loss = real_loss + fake_loss

    if mode == "min_fake":
        g_loss = - fake_loss
    elif mode == "max_real":
        g_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(
            logits=fake_logits, labels=tf.ones_like(fake_logits)))
    else:
        raise ValueError("Unknown mode: %s. Only 'min_fake' and 'max_real' "
                         "are allowed.")

    return g_loss, d_loss



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various classifier classes.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=not-context-manager, too-many-arguments, too-many-locals

import tensorflow as tf

from texar.utils.exceptions import TexarError
from texar.modules.classifiers.classifier_base import ClassifierBase
from texar.modules.encoders.conv_encoders import Conv1DEncoder
from texar.utils import utils
from texar.hyperparams import HParams

__all__ = [
    "Conv1DClassifier"
]

class Conv1DClassifier(ClassifierBase):
    """Simple Conv-1D classifier.
    This is a combination of the
    :class:`~texar.modules.Conv1DEncoder` with a classification layer.

    Args:
        hparams (dict, optional): Hyperparameters. Missing
            hyperparamerter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter sturcture and
            default values.

    Example:

        .. code-block:: python

            clas = Conv1DClassifier(hparams={'num_classes': 10})

            inputs = tf.random_uniform([64, 20, 256])
            logits, pred = clas(inputs)
            # logits == Tensor of shape [64, 10]
            # pred   == Tensor of shape [64]

    .. document private functions
    .. automethod:: _build
    """

    def __init__(self, hparams=None):
        ClassifierBase.__init__(self, hparams)

        with tf.variable_scope(self.variable_scope):
            encoder_hparams = utils.dict_fetch(
                hparams, Conv1DEncoder.default_hparams())
            self._encoder = Conv1DEncoder(hparams=encoder_hparams)

            # Add an additional dense layer if needed
            self._num_classes = self._hparams.num_classes
            if self._num_classes > 0:
                if self._hparams.num_dense_layers <= 0:
                    self._encoder.append_layer({"type": "Flatten"})

                logit_kwargs = self._hparams.logit_layer_kwargs
                if logit_kwargs is None:
                    logit_kwargs = {}
                elif not isinstance(logit_kwargs, HParams):
                    raise ValueError(
                        "hparams['logit_layer_kwargs'] must be a dict.")
                else:
                    logit_kwargs = logit_kwargs.todict()
                logit_kwargs.update({"units": self._num_classes})
                if 'name' not in logit_kwargs:
                    logit_kwargs['name'] = "logit_layer"

                self._encoder.append_layer(
                    {"type": "Dense", "kwargs": logit_kwargs})

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                # (1) Same hyperparameters as in Conv1DEncoder
                ...

                # (2) Additional hyperparameters
                "num_classes": 2,
                "logit_layer_kwargs": {
                    "use_bias": False
                },
                "name": "conv1d_classifier"
            }

        Here:

        1. Same hyperparameters as in :class:`~texar.modules.Conv1DEncoder`.
        See the :meth:`~texar.modules.Conv1DEncoder.default_hparams`.
        An instance of Conv1DEncoder is created for feature extraction.

        2. Additional hyperparameters:

            "num_classes" : int
                Number of classes:

                - If **`> 0`**, an additional :tf_main:`Dense <layers/Dense>` \
                layer is appended to the encoder to compute the logits over \
                classes.
                - If **`<= 0`**, no dense layer is appended. The number of \
                classes is assumed to be the final dense layer size of the \
                encoder.

            "logit_layer_kwargs" : dict
                Keyword arguments for the logit Dense layer constructor,
                except for argument "units" which is set to "num_classes".
                Ignored if no extra logit layer is appended.

            "name" : str
                Name of the classifier.
        """
        hparams = Conv1DEncoder.default_hparams()
        hparams.update({
            "name": "conv1d_classifier",
            "num_classes": 2, #set to <=0 to avoid appending output layer
            "logit_layer_kwargs": {"use_bias": False}
        })
        return hparams

    def _build(self,    # pylint: disable=arguments-differ
               inputs,
               sequence_length=None,
               dtype=None,
               mode=None):
        """Feeds the inputs through the network and makes classification.

        The arguments are the same as in :class:`~texar.modules.Conv1DEncoder`.

        The predictions of binary classification ("num_classes"=1) and
        multi-way classification ("num_classes">1) are different, as explained
        below.

        Args:
            inputs: The inputs to the network, which is a 3D tensor. See
                :class:`~texar.modules.Conv1DEncoder` for more details.
            sequence_length (optional): An int tensor of shape `[batch_size]`
                containing the length of each element in :attr:`inputs`.
                If given, time steps beyond the length will first be masked out
                before feeding to the layers.
            dtype (optional): Type of the inputs. If not provided, infers
                from inputs automatically.
            mode (optional): A tensor taking value in
                :tf_main:`tf.estimator.ModeKeys <estimator/ModeKeys>`, including
                `TRAIN`, `EVAL`, and `PREDICT`. If `None`,
                :func:`texar.global_mode` is used.

        Returns:
            A tuple `(logits, pred)`, where

            - **`logits`** is a Tensor of shape `[batch_size, num_classes]`\
            for `num_classes` >1, and `[batch_size]` for `num_classes` =1 \
            (i.e., binary classification).
            - **`pred`** is the prediction, a Tensor of shape `[batch_size]` \
            and type `tf.int64`. For binary classification, the standard \
            sigmoid function is used for prediction, and the class labels are \
            `{0, 1}`.
        """
        logits = self._encoder(inputs, sequence_length, dtype, mode)

        num_classes = self._hparams.num_classes
        is_binary = num_classes == 1
        is_binary = is_binary or (num_classes <= 0 and logits.shape[1] == 1)

        if is_binary:
            pred = tf.greater(logits, 0)
            logits = tf.reshape(logits, [-1])
        else:
            pred = tf.argmax(logits, 1)
        pred = tf.to_int64(tf.reshape(pred, [-1]))

        self._built = True

        return logits, pred

    @property
    def trainable_variables(self):
        """The list of trainable variables of the module.
        """
        if not self._built:
            raise TexarError(
                "Attempting to access trainable_variables before module %s "
                "was fully built. The module is built once it is called, "
                "e.g., with `%s(...)`" % (self.name, self.name))
        return self._encoder.trainable_variables

    @property
    def num_classes(self):
        """The number of classes.
        """
        return self._num_classes

    @property
    def nn(self): # pylint: disable=invalid-name
        """The classifier neural network.
        """
        return self._encoder

    def has_layer(self, layer_name):
        """Returns `True` if the network with the name exists. Returns `False`
        otherwise.

        Args:
            layer_name (str): Name of the layer.
        """
        return self._encoder.has_layer(layer_name)

    def layer_by_name(self, layer_name):
        """Returns the layer with the name. Returns 'None' if the layer name
        does not exist.

        Args:
            layer_name (str): Name of the layer.
        """
        return self._encoder.layer_by_name(layer_name)

    @property
    def layers_by_name(self):
        """A dictionary mapping layer names to the layers.
        """
        return self._encoder.layers_by_name

    @property
    def layers(self):
        """A list of the layers.
        """
        return self._encoder.layers

    @property
    def layer_names(self):
        """A list of uniquified layer names.
        """
        return self._encoder.layer_names

    def layer_outputs_by_name(self, layer_name):
        """Returns the output tensors of the layer with the specified name.
        Returns `None` if the layer name does not exist.

        Args:
            layer_name (str): Name of the layer.
        """
        return self._encoder.layer_outputs_by_name(layer_name)

    @property
    def layer_outputs(self):
        """A list containing output tensors of each layer.
        """
        return self._encoder.layer_outputs



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various RNN decoders.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=no-name-in-module, too-many-arguments, too-many-locals
# pylint: disable=not-context-manager, protected-access, invalid-name

import collections
import copy

import tensorflow as tf
from tensorflow.contrib.seq2seq import AttentionWrapper
from tensorflow.python.util import nest
from tensorflow.contrib.seq2seq import tile_batch

from texar.modules.decoders.rnn_decoder_base import RNNDecoderBase
from texar.utils import utils

__all__ = [
    "BasicRNNDecoderOutput",
    "AttentionRNNDecoderOutput",
    "BasicRNNDecoder",
    "AttentionRNNDecoder"
]


class BasicRNNDecoderOutput(
        collections.namedtuple("BasicRNNDecoderOutput",
                               ("logits", "sample_id", "cell_output"))):
    """The outputs of basic RNN decoder that include both RNN outputs and
    sampled ids at each step. This is also used to store results of all the
    steps after decoding the whole sequence.

    Attributes:
        logits: The outputs of RNN (at each step/of all steps) by applying the
            output layer on cell outputs. E.g., in
            :class:`~texar.modules.BasicRNNDecoder` with default
            hyperparameters, this is a Tensor of
            shape `[batch_size, max_time, vocab_size]` after decoding the
            whole sequence.
        sample_id: The sampled results (at each step/of all steps). E.g., in
            BasicRNNDecoder with decoding strategy of train_greedy,
            this is a Tensor
            of shape `[batch_size, max_time]` containing the sampled token
            indexes of all steps.
        cell_output: The output of RNN cell (at each step/of all steps).
            This is the results prior to the output layer. E.g., in
            BasicRNNDecoder with default
            hyperparameters, this is a Tensor of
            shape `[batch_size, max_time, cell_output_size]` after decoding
            the whole sequence.
    """
    pass

class AttentionRNNDecoderOutput(
        collections.namedtuple(
            "AttentionRNNDecoderOutput",
            ["logits", "sample_id", "cell_output",
             "attention_scores", "attention_context"])):
    """The outputs of attention RNN decoders that additionally include
    attention results.

    Attributes:
        logits: The outputs of RNN (at each step/of all steps) by applying the
            output layer on cell outputs. E.g., in
            :class:`~texar.modules.AttentionRNNDecoder`, this is a Tensor of
            shape `[batch_size, max_time, vocab_size]` after decoding.
        sample_id: The sampled results (at each step/of all steps). E.g., in
            :class:`~texar.modules.AttentionRNNDecoder` with decoding strategy
            of train_greedy, this
            is a Tensor of shape `[batch_size, max_time]` containing the
            sampled token indexes of all steps.
        cell_output: The output of RNN cell (at each step/of all steps).
            This is the results prior to the output layer. E.g., in
            AttentionRNNDecoder with default
            hyperparameters, this is a Tensor of
            shape `[batch_size, max_time, cell_output_size]` after decoding
            the whole sequence.
        attention_scores: A single or tuple of `Tensor`(s) containing the
            alignments emitted (at the previous time step/of all time steps)
            for each attention mechanism.
        attention_context: The attention emitted (at the previous time step/of
            all time steps).
    """
    pass


class BasicRNNDecoder(RNNDecoderBase):
    """Basic RNN decoder.

    Args:
        cell (RNNCell, optional): An instance of
            :tf_main:`RNNCell <ontrib/rnn/RNNCell>`. If `None`
            (default), a cell is created as specified in
            :attr:`hparams`.
        cell_dropout_mode (optional): A Tensor taking value of
            :tf_main:`tf.estimator.ModeKeys <estimator/ModeKeys>`, which
            toggles dropout in the RNN cell (e.g., activates dropout in
            TRAIN mode). If `None`, :func:`~texar.global_mode` is used.
            Ignored if :attr:`cell` is given.
        vocab_size (int, optional): Vocabulary size. Required if
            :attr:`output_layer` is `None`.
        output_layer (optional): An output layer that transforms cell output
            to logits. This can be:

            - A callable layer, e.g., an instance \
            of :tf_main:`tf.layers.Layer <layers/Layer>`.
            - A tensor. A dense layer will be created using the tensor \
            as the kernel weights. The bias of the dense layer is determined by\
            `hparams.output_layer_bias`. This can be used to tie the output \
            layer with the input embedding matrix, as proposed in \
            https://arxiv.org/pdf/1608.05859.pdf
            - `None`. A dense layer will be created based on attr:`vocab_size`\
            and `hparams.output_layer_bias`.
            - If no output layer after the cell output is needed, set \
            `(vocab_size=None, output_layer=tf.identity)`.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparamerter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter sturcture and
            default values.

    See :meth:`~texar.modules.RNNDecoderBase._build` for the inputs and outputs
    of the decoder. The decoder returns
    `(outputs, final_state, sequence_lengths)`, where `outputs` is an instance
    of :class:`~texar.modules.BasicRNNDecoderOutput`.

    Example:

        .. code-block:: python

            embedder = WordEmbedder(vocab_size=data.vocab.size)
            decoder = BasicRNNDecoder(vocab_size=data.vocab.size)

            # Training loss
            outputs, _, _ = decoder(
                decoding_strategy='train_greedy',
                inputs=embedder(data_batch['text_ids']),
                sequence_length=data_batch['length']-1)

            loss = tx.losses.sequence_sparse_softmax_cross_entropy(
                labels=data_batch['text_ids'][:, 1:],
                logits=outputs.logits,
                sequence_length=data_batch['length']-1)

            # Inference sample
            outputs, _, _ = decoder(
                decoding_strategy='infer_sample',
                start_tokens=[data.vocab.bos_token_id]*100,
                end_token=data.vocab.eos.token_id,
                embedding=embedder,
                max_decoding_length=60,
                mode=tf.estimator.ModeKeys.PREDICT)

            sample_id = sess.run(outputs.sample_id)
            sample_text = tx.utils.map_ids_to_strs(sample_id, data.vocab)
            print(sample_text)
            # [
            #   the first sequence sample .
            #   the second sequence sample .
            #   ...
            # ]
    """

    def __init__(self,
                 cell=None,
                 cell_dropout_mode=None,
                 vocab_size=None,
                 output_layer=None,
                 hparams=None):
        RNNDecoderBase.__init__(
            self, cell, vocab_size, output_layer, cell_dropout_mode, hparams)

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "rnn_cell": default_rnn_cell_hparams(),
                "max_decoding_length_train": None,
                "max_decoding_length_infer": None,
                "helper_train": {
                    "type": "TrainingHelper",
                    "kwargs": {}
                }
                "helper_infer": {
                    "type": "SampleEmbeddingHelper",
                    "kwargs": {}
                }
                "name": "basic_rnn_decoder"
            }

        Here:

        "rnn_cell" : dict
            A dictionary of RNN cell hyperparameters. Ignored if
            :attr:`cell` is given to the decoder constructor.
            The default value is defined in
            :func:`~texar.core.default_rnn_cell_hparams`.

        "max_decoding_length_train": int or None
            Maximum allowed number of decoding steps in training mode.
            If `None` (default), decoding is
            performed until fully done, e.g., encountering the <EOS> token.
            Ignored if `max_decoding_length` is given when calling
            the decoder.

        "max_decoding_length_infer" : int or None
            Same as "max_decoding_length_train" but for inference mode.

        "helper_train" : dict
            The hyperparameters of the helper used in training.
            "type" can be a helper class, its name or module path, or a
            helper instance. If a class name is given, the class must be
            from module :tf_main:`tf.contrib.seq2seq <contrib/seq2seq>`,
            :mod:`texar.modules`, or :mod:`texar.custom`. This is used
            only when both `decoding_strategy` and `helper` augments are
            `None` when calling the decoder. See
            :meth:`~texar.modules.RNNDecoderBase._build` for more details.

        "helper_infer": dict
            Same as "helper_train" but during inference mode.

        "name" : str
            Name of the decoder.

            The default value is "basic_rnn_decoder".
        """
        hparams = RNNDecoderBase.default_hparams()
        hparams["name"] = "basic_rnn_decoder"
        return hparams

    def initialize(self, name=None):
        return self._helper.initialize() + (self._initial_state,)

    def step(self, time, inputs, state, name=None):
        cell_outputs, cell_state = self._cell(inputs, state)
        logits = self._output_layer(cell_outputs)
        sample_ids = self._helper.sample(
            time=time, outputs=logits, state=cell_state)
        reach_max_time = tf.equal(time+1, self.max_decoding_length)

        (finished, next_inputs, next_state) = self._helper.next_inputs(
            time=time,
            outputs=logits,
            state=cell_state,
            sample_ids=sample_ids,
            reach_max_time=reach_max_time)

        outputs = BasicRNNDecoderOutput(logits, sample_ids, cell_outputs)
        return (outputs, next_state, next_inputs, finished)

    def finalize(self, outputs, final_state, sequence_lengths):
        return outputs, final_state

    @property
    def output_size(self):
        """Output size of one step.
        """
        return BasicRNNDecoderOutput(
            logits=self._rnn_output_size(),
            sample_id=self._helper.sample_ids_shape,
            cell_output=self._cell.output_size)

    @property
    def output_dtype(self):
        """Types of output of one step.
        """
        # Assume the dtype of the cell is the output_size structure
        # containing the input_state's first component's dtype.
        # Return that structure and the sample_ids_dtype from the helper.
        dtype = nest.flatten(self._initial_state)[0].dtype
        return BasicRNNDecoderOutput(
            logits=nest.map_structure(lambda _: dtype, self._rnn_output_size()),
            sample_id=self._helper.sample_ids_dtype,
            cell_output=nest.map_structure(
                lambda _: dtype, self._cell.output_size))


class AttentionRNNDecoder(RNNDecoderBase):
    """RNN decoder with attention mechanism.

    Args:
        memory: The memory to query, e.g., the output of an RNN encoder. This
            tensor should be shaped `[batch_size, max_time, dim]`.
        memory_sequence_length (optional): A tensor of shape `[batch_size]`
            containing the sequence lengths for the batch
            entries in memory. If provided, the memory tensor rows are masked
            with zeros for values past the respective sequence lengths.
        cell (RNNCell, optional): An instance of `RNNCell`. If `None`, a cell
            is created as specified in :attr:`hparams`.
        cell_dropout_mode (optional): A Tensor taking value of
            :tf_main:`tf.estimator.ModeKeys <estimator/ModeKeys>`, which
            toggles dropout in the RNN cell (e.g., activates dropout in
            TRAIN mode). If `None`, :func:`~texar.global_mode` is used.
            Ignored if :attr:`cell` is given.
        vocab_size (int, optional): Vocabulary size. Required if
            :attr:`output_layer` is `None`.
        output_layer (optional): An output layer that transforms cell output
            to logits. This can be:

            - A callable layer, e.g., an instance \
            of :tf_main:`tf.layers.Layer <layers/Layer>`.
            - A tensor. A dense layer will be created using the tensor \
            as the kernel weights. The bias of the dense layer is determined by\
            `hparams.output_layer_bias`. This can be used to tie the output \
            layer with the input embedding matrix, as proposed in \
            https://arxiv.org/pdf/1608.05859.pdf
            - `None`. A dense layer will be created based on attr:`vocab_size`\
            and `hparams.output_layer_bias`.
            - If no output layer after the cell output is needed, set \
            `(vocab_size=None, output_layer=tf.identity)`.
        cell_input_fn (callable, optional): A callable that produces RNN cell
            inputs. If `None` (default), the default is used:
            `lambda inputs, attention: tf.concat([inputs, attention], -1)`,
            which cancats regular RNN cell inputs with attentions.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparamerter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter sturcture and
            default values.

    See :meth:`~texar.modules.RNNDecoderBase._build` for the inputs and outputs
    of the decoder. The decoder returns
    `(outputs, final_state, sequence_lengths)`, where `outputs` is an instance
    of :class:`~texar.modules.AttentionRNNDecoderOutput`.

    Example:

        .. code-block:: python

            # Encodes the source
            enc_embedder = WordEmbedder(data.source_vocab.size, ...)
            encoder = UnidirectionalRNNEncoder(...)

            enc_outputs, _ = encoder(
                inputs=enc_embedder(data_batch['source_text_ids']),
                sequence_length=data_batch['source_length'])

            # Decodes while attending to the source
            dec_embedder = WordEmbedder(vocab_size=data.target_vocab.size, ...)
            decoder = AttentionRNNDecoder(
                memory=enc_outputs,
                memory_sequence_length=data_batch['source_length'],
                vocab_size=data.target_vocab.size)

            outputs, _, _ = decoder(
                decoding_strategy='train_greedy',
                inputs=dec_embedder(data_batch['target_text_ids']),
                sequence_length=data_batch['target_length']-1)
    """
    def __init__(self,
                 memory,
                 memory_sequence_length=None,
                 cell=None,
                 cell_dropout_mode=None,
                 vocab_size=None,
                 output_layer=None,
                 #attention_layer=None, # TODO(zhiting): only valid for tf>=1.0
                 cell_input_fn=None,
                 hparams=None):
        RNNDecoderBase.__init__(
            self, cell, vocab_size, output_layer, cell_dropout_mode, hparams)

        attn_hparams = self._hparams['attention']
        attn_kwargs = attn_hparams['kwargs'].todict()

        # Parse the 'probability_fn' argument
        if 'probability_fn' in attn_kwargs:
            prob_fn = attn_kwargs['probability_fn']
            if prob_fn is not None and not callable(prob_fn):
                prob_fn = utils.get_function(
                    prob_fn,
                    ['tensorflow.nn', 'tensorflow.contrib.sparsemax',
                     'tensorflow.contrib.seq2seq'])
            attn_kwargs['probability_fn'] = prob_fn

        attn_kwargs.update({
            "memory_sequence_length": memory_sequence_length,
            "memory": memory})
        self._attn_kwargs = attn_kwargs
        attn_modules = ['tensorflow.contrib.seq2seq', 'texar.custom']
        # Use variable_scope to ensure all trainable variables created in
        # the attention mechanism are collected
        with tf.variable_scope(self.variable_scope):
            attention_mechanism = utils.check_or_get_instance(
                attn_hparams["type"], attn_kwargs, attn_modules,
                classtype=tf.contrib.seq2seq.AttentionMechanism)

        self._attn_cell_kwargs = {
            "attention_layer_size": attn_hparams["attention_layer_size"],
            "alignment_history": attn_hparams["alignment_history"],
            "output_attention": attn_hparams["output_attention"],
        }
        self._cell_input_fn = cell_input_fn
        # Use variable_scope to ensure all trainable variables created in
        # AttentionWrapper are collected
        with tf.variable_scope(self.variable_scope):
            #if attention_layer is not None:
            #    self._attn_cell_kwargs["attention_layer_size"] = None
            attn_cell = AttentionWrapper(
                self._cell,
                attention_mechanism,
                cell_input_fn=self._cell_input_fn,
                #attention_layer=attention_layer,
                **self._attn_cell_kwargs)
            self._cell = attn_cell

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values:

        Common hyperparameters are the same as in
        :class:`~texar.modules.BasicRNNDecoder`.
        :meth:`~texar.modules.BasicRNNDecoder.default_hparams`.
        Additional hyperparameters are for attention mechanism
        configuration.

        .. code-block:: python

            {
                "attention": {
                    "type": "LuongAttention",
                    "kwargs": {
                        "num_units": 256,
                    },
                    "attention_layer_size": None,
                    "alignment_history": False,
                    "output_attention": True,
                },
                # The following hyperparameters are the same as with
                # `BasicRNNDecoder`
                "rnn_cell": default_rnn_cell_hparams(),
                "max_decoding_length_train": None,
                "max_decoding_length_infer": None,
                "helper_train": {
                    "type": "TrainingHelper",
                    "kwargs": {}
                }
                "helper_infer": {
                    "type": "SampleEmbeddingHelper",
                    "kwargs": {}
                }
                "name": "attention_rnn_decoder"
            }

        Here:

        "attention" : dict
            Attention hyperparameters, including:

            "type" : str or class or instance
                The attention type. Can be an attention class, its name or
                module path, or a class instance. The class must be a subclass
                of :tf_main:`TF AttentionMechanism
                <contrib/seq2seq/AttentionMechanism>`. If class name is
                given, the class must be from modules
                :tf_main:`tf.contrib.seq2seq <contrib/seq2seq>` or
                :mod:`texar.custom`.

                Example:

                    .. code-block:: python

                        # class name
                        "type": "LuongAttention"
                        "type": "BahdanauAttention"
                        # module path
                        "type": "tf.contrib.seq2seq.BahdanauMonotonicAttention"
                        "type": "my_module.MyAttentionMechanismClass"
                        # class
                        "type": tf.contrib.seq2seq.LuongMonotonicAttention
                        # instance
                        "type": LuongAttention(...)

            "kwargs" : dict
                keyword arguments for the attention class constructor.
                Arguments :attr:`memory` and
                :attr:`memory_sequence_length` should **not** be
                specified here because they are given to the decoder
                constructor. Ignored if "type" is an attention class
                instance. For example

                Example:

                    .. code-block:: python

                        "type": "LuongAttention",
                        "kwargs": {
                            "num_units": 256,
                            "probability_fn": tf.nn.softmax
                        }

                    Here "probability_fn" can also be set to the string name
                    or module path to a probability function.

                "attention_layer_size" : int or None
                    The depth of the attention (output) layer. The context and
                    cell output are fed into the attention layer to generate
                    attention at each time step.
                    If `None` (default), use the context as attention at each
                    time step.

                "alignment_history": bool
                    whether to store alignment history from all time steps
                    in the final output state. (Stored as a time major
                    `TensorArray` on which you must call `stack()`.)

                "output_attention": bool
                    If `True` (default), the output at each time step is
                    the attention value. This is the behavior of Luong-style
                    attention mechanisms. If `False`, the output at each
                    time step is the output of `cell`.  This is the
                    beahvior of Bhadanau-style attention mechanisms.
                    In both cases, the `attention` tensor is propagated to
                    the next time step via the state and is used there.
                    This flag only controls whether the attention mechanism
                    is propagated up to the next cell in an RNN stack or to
                    the top RNN output.
        """
        hparams = RNNDecoderBase.default_hparams()
        hparams["name"] = "attention_rnn_decoder"
        hparams["attention"] = {
            "type": "LuongAttention",
            "kwargs": {
                "num_units": 256,
            },
            "attention_layer_size": None,
            "alignment_history": False,
            "output_attention": True,
        }
        return hparams

    # pylint: disable=arguments-differ
    def _get_beam_search_cell(self, beam_width):
        """Returns the RNN cell for beam search decoding.
        """
        with tf.variable_scope(self.variable_scope, reuse=True):
            attn_kwargs = copy.copy(self._attn_kwargs)

            memory = attn_kwargs['memory']
            attn_kwargs['memory'] = tile_batch(memory, multiplier=beam_width)

            memory_seq_length = attn_kwargs['memory_sequence_length']
            if memory_seq_length is not None:
                attn_kwargs['memory_sequence_length'] = tile_batch(
                    memory_seq_length, beam_width)

            attn_modules = ['tensorflow.contrib.seq2seq', 'texar.custom']
            bs_attention_mechanism = utils.check_or_get_instance(
                self._hparams.attention.type, attn_kwargs, attn_modules,
                classtype=tf.contrib.seq2seq.AttentionMechanism)

            bs_attn_cell = AttentionWrapper(
                self._cell._cell,
                bs_attention_mechanism,
                cell_input_fn=self._cell_input_fn,
                **self._attn_cell_kwargs)

            self._beam_search_cell = bs_attn_cell

            return bs_attn_cell

    def initialize(self, name=None):
        helper_init = self._helper.initialize()

        flat_initial_state = nest.flatten(self._initial_state)
        dtype = flat_initial_state[0].dtype
        initial_state = self._cell.zero_state(
            batch_size=tf.shape(flat_initial_state[0])[0], dtype=dtype)
        initial_state = initial_state.clone(cell_state=self._initial_state)

        return [helper_init[0], helper_init[1], initial_state]

    def step(self, time, inputs, state, name=None):
        wrapper_outputs, wrapper_state = self._cell(inputs, state)
        # Essentisally the same as in BasicRNNDecoder.step()
        logits = self._output_layer(wrapper_outputs)
        sample_ids = self._helper.sample(
            time=time, outputs=logits, state=wrapper_state)
        reach_max_time = tf.equal(time+1, self.max_decoding_length)

        (finished, next_inputs, next_state) = self._helper.next_inputs(
            time=time,
            outputs=logits,
            state=wrapper_state,
            sample_ids=sample_ids,
            reach_max_time=reach_max_time)

        attention_scores = wrapper_state.alignments
        attention_context = wrapper_state.attention
        outputs = AttentionRNNDecoderOutput(
            logits, sample_ids, wrapper_outputs,
            attention_scores, attention_context)

        return (outputs, next_state, next_inputs, finished)

    def finalize(self, outputs, final_state, sequence_lengths):
        return outputs, final_state

    def _alignments_size(self):
        # Reimplementation of the alignments_size of each of
        # AttentionWrapper.attention_mechanisms. The original implementation
        # of `_BaseAttentionMechanism._alignments_size`:
        #
        #    self._alignments_size = (self._keys.shape[1].value or
        #                       array_ops.shape(self._keys)[1])
        #
        # can be `None` when the seq length of encoder outputs are priori
        # unknown.
        alignments_size = []
        for am in self._cell._attention_mechanisms:
            az = (am._keys.shape[1].value or tf.shape(am._keys)[1:-1])
            alignments_size.append(az)
        return self._cell._item_or_tuple(alignments_size)

    @property
    def output_size(self):
        return AttentionRNNDecoderOutput(
            logits=self._rnn_output_size(),
            sample_id=self._helper.sample_ids_shape,
            cell_output=self._cell.output_size,
            attention_scores=self._alignments_size(),
            attention_context=self._cell.state_size.attention)

    @property
    def output_dtype(self):
        """Types of output of one step.
        """
        # Assume the dtype of the cell is the output_size structure
        # containing the input_state's first component's dtype.
        # Return that structure and the sample_ids_dtype from the helper.
        dtype = nest.flatten(self._initial_state)[0].dtype
        return AttentionRNNDecoderOutput(
            logits=nest.map_structure(lambda _: dtype, self._rnn_output_size()),
            sample_id=self._helper.sample_ids_dtype,
            cell_output=nest.map_structure(
                lambda _: dtype, self._cell.output_size),
            attention_scores=nest.map_structure(
                lambda _: dtype, self._alignments_size()),
            attention_context=nest.map_structure(
                lambda _: dtype, self._cell.state_size.attention))

    def zero_state(self, batch_size, dtype):
        """Returns zero state of the basic cell.
        Equivalent to :attr:`decoder.cell._cell.zero_state`.
        """
        return self._cell._cell.zero_state(batch_size=batch_size, dtype=dtype)

    def wrapper_zero_state(self, batch_size, dtype):
        """Returns zero state of the attention-wrapped cell.
        Equivalent to :attr:`decoder.cell.zero_state`.
        """
        return self._cell.zero_state(batch_size=batch_size, dtype=dtype)

    @property
    def state_size(self):
        """The state size of the basic cell.
        Equivalent to :attr:`decoder.cell._cell.state_size`.
        """
        return self._cell._cell.state_size

    @property
    def wrapper_state_size(self):
        """The state size of the attention-wrapped cell.
        Equivalent to :attr:`decoder.cell.state_size`.
        """
        return self._cell.state_size




# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various convolutional network encoders.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from texar.modules.encoders.encoder_base import EncoderBase
from texar.modules.networks.conv_networks import Conv1DNetwork

__all__ = [
    "Conv1DEncoder"
]

class Conv1DEncoder(Conv1DNetwork, EncoderBase):
    """Simple Conv-1D encoder which consists of a sequence of conv layers
    followed with a sequence of dense layers.

    Wraps :class:`~texar.modules.Conv1DNetwork` to be a subclass of
    :class:`~texar.modules.EncoderBase`. Has exact the same functionality
    with :class:`~texar.modules.Conv1DNetwork`.
    """

    def __init__(self, hparams=None): # pylint: disable=super-init-not-called
        Conv1DNetwork.__init__(self, hparams)

    @staticmethod
    def default_hparams():
        """Returns a dictionary of hyperparameters with default values.

        The same as :meth:`~texar.modules.Conv1DNetwork.default_hparams`
        of :class:`~texar.modules.Conv1DNetwork`, except that the default name
        is 'conv_encoder'.
        """
        hparams = Conv1DNetwork.default_hparams()
        hparams['name'] = 'conv_encoder'
        return hparams

